"""Data processing unit tests."""

# Native imports
from logging import getLogger
from pathlib import Path

# Third-party imports
import numpy as np
import pandas as pd
import pytest

# Sea-Bird imports

# Internal imports
import seabirdscientific.instrument_data as idata
import seabirdscientific.processing as p
import seabirdscientific.conversion as c
from seabirdscientific.utils import close_enough, get_tolerance

test_data = Path("./tests/resources/test-data")
logger = getLogger(__name__)


class TestLowPassFilter:
    def test_low_pass_filter(self, request):
        expected_path = test_data / "SBE37SM-filtered.asc"
        expected = pd.read_csv(expected_path)["Tv290C"].values

        source_path = test_data / "SBE37SM-unfiltered.asc"
        source = pd.read_csv(source_path)["Tv290C"].values

        filtered = p.low_pass_filter(source, 10000, sample_interval=120)

        request.node.return_value = filtered.tolist()

        assert np.allclose(expected, filtered, rtol=0, atol=1e-4)


class TestAlignCtdFathomCases:
    # Test align CTD with same test data as Fathom
    def test_align_ctd_basic_forward(self, request):
        badFlagValue = -9.99e-29

        align_cases = [
            # Basic forward align cases
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),  # original
                15,  # offset
                15,  # sample interval
                np.array([15, 30, 45, 60, 75, 90, 105, badFlagValue]),  # expected
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                10,
                15,
                np.array([10, 25, 40, 55, 70, 85, 100, badFlagValue]),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                5,
                15,
                np.array([5, 20, 35, 50, 65, 80, 95, badFlagValue]),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                0,
                15,
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                11.575,
                15,
                np.array([11.575, 26.575, 41.575, 56.575, 71.575, 86.575, 101.575, badFlagValue]),
            ],
            # Advanced forward align cases
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                30,
                15,
                np.array([30, 45, 60, 75, 90, 105, badFlagValue, badFlagValue]),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                20,
                15,
                np.array([20, 35, 50, 65, 80, 95, badFlagValue, badFlagValue]),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                106,
                15,
                np.array([badFlagValue] * 8),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                60,
                15,
                np.array(
                    [60, 75, 90, 105, badFlagValue, badFlagValue, badFlagValue, badFlagValue]
                ),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                130,
                15,
                np.array([badFlagValue] * 8),
            ],
            # Basic backwards align cases
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                -15,
                15,
                np.array([badFlagValue, 0, 15, 30, 45, 60, 75, 90]),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                -10,
                15,
                np.array([badFlagValue, 5, 20, 35, 50, 65, 80, 95]),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                -5,
                15,
                np.array([badFlagValue, 10, 25, 40, 55, 70, 85, 100]),
            ],
            # Advanced backwards align cases
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                -30,
                15,
                np.array([badFlagValue, badFlagValue, 0, 15, 30, 45, 60, 75]),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                -20,
                15,
                np.array([badFlagValue, badFlagValue, 10, 25, 40, 55, 70, 85]),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                -106,
                15,
                np.array([badFlagValue] * 8),
            ],
            [
                np.array([0, 15, 30, 45, 60, 75, 90, 105]),
                -130,
                15,
                np.array([badFlagValue] * 8),
            ],
        ]

        for case in align_cases:
            result = p.align_ctd(case[0], case[1], case[2])
            assert np.array_equal(case[3], result)


class TestAlignCtd:
    expected_data_path = test_data / "SBE37SM-align.cnv"
    expected_data = idata.cnv_to_instrument_data(expected_data_path)
    # expected data was algined with:
    #   sample_interval=120,
    #   tv290C + 12s
    #   cond0S/m + 150s
    #   prdM + 120s
    #   prdE + 0s
    #   sal00 - 240s

    source_data_path = test_data / "SBE37SM.cnv"
    source_data = idata.cnv_to_instrument_data(source_data_path)

    def test_align_ctd_add_simple_pass(self, request):
        # Fix last valid sample
        self.expected_data.measurements["tv290C"].values[
            len(self.expected_data.measurements["tv290C"].values) - 1
        ] = -9.99e-29
        result = p.align_ctd(self.source_data.measurements["tv290C"].values, 12, 120)
        request.node.return_value = result.tolist()
        assert np.allclose(self.expected_data.measurements["tv290C"].values, result, atol=0.0001)

    def test_align_ctd_add_simple_pass_2(self, request):
        # Fix last valid sample
        self.expected_data.measurements["cond0S/m"].values[
            len(self.expected_data.measurements["cond0S/m"].values) - 2
        ] = -9.99e-29
        result = p.align_ctd(self.source_data.measurements["cond0S/m"].values, 150, 120)
        request.node.return_value = result.tolist()
        assert np.allclose(self.expected_data.measurements["cond0S/m"].values, result, atol=0.0001)

    def test_align_ctd_add_exact_factor(self, request):
        result = p.align_ctd(self.source_data.measurements["prdM"].values, 120, 120)
        request.node.return_value = result.tolist()
        assert np.allclose(self.expected_data.measurements["prdM"].values, result, atol=0.0001)

    def test_align_ctd_no_change(self, request):
        result = p.align_ctd(self.source_data.measurements["prdE"].values, 0, 120)
        request.node.return_value = result.tolist()
        assert np.allclose(self.expected_data.measurements["prdE"].values, result, atol=0.0001)

    def test_align_ctd_subtract(self, request):
        result = p.align_ctd(self.source_data.measurements["sal00"].values, -240, 120)
        request.node.return_value = result.tolist()
        assert np.allclose(self.expected_data.measurements["sal00"].values, result, atol=0.0001)


class TestCellThermalMass:
    def test_cell_thermal_mass_pass(self, request):
        expected_data_path = test_data / "SBE37SM-ctm.cnv"
        expected_data = idata.cnv_to_instrument_data(expected_data_path)
        source_data_path = test_data / "SBE37SM.cnv"
        source_data = idata.cnv_to_instrument_data(source_data_path)
        corrected_conductivity = p.cell_thermal_mass(
            source_data.measurements["tv290C"].values,
            source_data.measurements["cond0S/m"].values,
            0.03,
            7.0,
            120.0,
        )
        request.node.return_value = corrected_conductivity.tolist()
        assert np.allclose(
            expected_data.measurements["cond0S/m"].values,
            corrected_conductivity,
            atol=0.000001,
        )


class TestLoopEdit:
    def test_loop_edit_pressure_min_velocity_pass(self, request):
        expected_data = idata.cnv_to_instrument_data(
            test_data / "CAST0002_mod_filt_loop_min_v.cnv"
        )
        data = idata.cnv_to_instrument_data(test_data / "CAST0002_mod_filt.cnv")

        p.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=p.MinVelocityType.FIXED,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=False,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        request.node.return_value = result_flags.tolist()
        assert mismatches / len(result_flags) < 0.01

    def test_loop_edit_pressure_min_velocity_remove_soak_pass(self, request):
        expected_data = idata.cnv_to_instrument_data(
            test_data / "CAST0002_mod_filt_loop_min_v_remove_soak.cnv"
        )
        data = idata.cnv_to_instrument_data(test_data / "CAST0002_mod_filt.cnv")

        p.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=p.MinVelocityType.FIXED,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=True,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        request.node.return_value = result_flags.tolist()
        assert mismatches / len(result_flags) < 0.01

    def test_loop_edit_pressure_min_velocity_reset_flags_pass(self, request):
        expected_data = idata.cnv_to_instrument_data(
            test_data / "CAST0002_mod_filt_loop_min_v.cnv"
        )
        data = idata.cnv_to_instrument_data(
            test_data / "CAST0002_mod_filt_loop_min_v_remove_soak.cnv"
        )

        p.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=p.MinVelocityType.FIXED,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=False,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        request.node.return_value = result_flags.tolist()
        assert mismatches / len(result_flags) < 0.01

    def test_loop_edit_pressure_min_velocity_exclude_flags_pass(self, request):
        expected_data = idata.cnv_to_instrument_data(
            test_data / "CAST0002_mod_filt_loop_min_v_remove_soak.cnv"
        )
        data = idata.cnv_to_instrument_data(
            test_data / "CAST0002_mod_filt_loop_min_v_exclude_flags_from_remove_soak.cnv"
        )

        p.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=p.MinVelocityType.FIXED,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=False,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=True,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        request.node.return_value = result_flags.tolist()
        assert mismatches / len(result_flags) < 0.01

    def test_loop_edit_pressure_mean_speed_percent_remove_soak_pass(self, request):
        expected_data = idata.cnv_to_instrument_data(
            test_data / "CAST0002_mod_filt_loop_percent_remove_soak.cnv"
        )
        data = idata.cnv_to_instrument_data(test_data / "CAST0002_mod_filt.cnv")

        p.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=p.MinVelocityType.PERCENT,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=True,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        request.node.return_value = result_flags.tolist()
        # Less than 2% mismatched flags, outliers appear to conform to spec.
        assert mismatches / len(result_flags) < 0.02

    def test_loop_edit_pressure_mean_speed_percent_pass(self, request):
        expected_data = idata.cnv_to_instrument_data(
            test_data / "CAST0002_mod_filt_loop_percent.cnv"
        )
        data = idata.cnv_to_instrument_data(test_data / "CAST0002_mod_filt.cnv")

        p.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=p.MinVelocityType.PERCENT,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=False,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        # SeaSoft is including earliest samples as local maxima during the first
        # downcast which would be discarded when requiring a minimum soak depth
        expected_flags = expected_data.measurements["flag"].values[100:]
        result_flags = data.measurements["flag"].values[100:]
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        request.node.return_value = result_flags.tolist()
        # Less than 2% mismatched flags, outliers appear to conform to spec.
        assert mismatches / len(result_flags) < 0.02

    def test_loop_edit_pressure_min_velocity_pass_2(self, request):
        expected_data = idata.cnv_to_instrument_data(test_data / "SBE19plus_loop_edit.cnv")
        data = idata.cnv_to_instrument_data(test_data / "SBE19plus.cnv")

        p.loop_edit_pressure(
            pressure=data.measurements["prdM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=p.MinVelocityType.FIXED,
            min_velocity=0.25,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=True,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=True,
            exclude_flags=True,
        )

        # SeaSoft is including earliest samples as local maxima during the first
        # downcast which would be discarded when requiring a minimum soak depth
        expected_flags = expected_data.measurements["flag"].values[100:]
        result_flags = data.measurements["flag"].values[100:]
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        request.node.return_value = result_flags.tolist()
        # Less than 2% mismatched flags, outliers appear to conform to spec.
        assert mismatches / len(result_flags) < 0.02


class TestBinAverage:
    # fmt: off
    # I think the extra wide lines here are less annoying than the
    # extra tall lines formatted by black. Feel free to enable black
    # here if you disagree

    def test_bin_average_interp_bin_3(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_interp_bin3.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset=data,
            bin_variable='prdM',
            bin_size=3,
            interpolate=True
            )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_interp_bin_5(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_interp_bin5.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset=data,
            bin_variable='prdM',
            bin_size=5,
            interpolate=True
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_default(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_in2 = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped2.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg.cnv"
        source_out2 = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped2_binavg.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        data2 = idata.cnv_to_instrument_data(source_in2)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        expected2 = idata.cnv_to_instrument_data(source_out2)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
        )
        binavg2 = p.bin_average(
            dataset = data2,
            bin_variable = 'prdM',
            bin_size = 2,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)
            assert np.allclose(expected2[variable], binavg2[variable], rtol=0, atol=tolerance)

    def test_bin_average_include_surface(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_surface.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            include_surface_bin = True,
            surface_bin_min = 0,
            surface_bin_max = 1,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_interpolate(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_interp.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            interpolate = True,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_interpolate_surface(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_interp_surface.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            interpolate = True,
            include_surface_bin = True,
            surface_bin_min = 0,
            surface_bin_max = 1,
            surface_bin_value = 0.5,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_downcast(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_downcast.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            cast_type = p.CastType.DOWNCAST
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_downcast_interp(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_downcast_interp.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            cast_type = p.CastType.DOWNCAST,
            interpolate = True,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_downcast_interp_surface(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_downcast_interp_surface.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            cast_type = p.CastType.DOWNCAST,
            interpolate = True,
            include_surface_bin = True,
            surface_bin_min = 0,
            surface_bin_max = 1,
            surface_bin_value = 0.5,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_upcast(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_upcast.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            cast_type = p.CastType.UPCAST
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_upcast_interp(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_upcast_interp.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            cast_type = p.CastType.UPCAST,
            interpolate = True,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_upcast_interp_surface(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_upcast_interp_surface.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            cast_type = p.CastType.UPCAST,
            interpolate = True,
            include_surface_bin = True,
            surface_bin_min = 0,
            surface_bin_max = 1,
            surface_bin_value = 0.5,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_exclude_flags(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_wildedit.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_wildedit_binavg_exclude.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_include_flags(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_wildedit.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_wildedit_binavg_include.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            exclude_bad_scans = False
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_include_flags_interp(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_wildedit.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_wildedit_binavg_include_interp.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            interpolate = True,
            exclude_bad_scans = False,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_exclude_flags_interp(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_wildedit.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_wildedit_binavg_exclude_interp.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'prdM',
            bin_size = 2,
            interpolate = True,
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_time(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_time.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'timeS',
            bin_size = 10,
            cast_type = p.CastType.NA
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    def test_bin_average_scan(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_out = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_binavg_scan.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected = idata.cnv_to_instrument_data(source_out)._to_dataframe()
        binavg = p.bin_average(
            dataset = data,
            bin_variable = 'nScan',
            bin_size = 10,
            cast_type = p.CastType.NA
        )
        for variable in expected.columns:
            tolerance = get_tolerance(expected[variable])
            assert np.allclose(expected[variable], binavg[variable], rtol=0, atol=tolerance)

    # fmt: on


class TestWildEdit:
    def test_wild_edit_pass(self, request):
        expected_dataset = idata.cnv_to_instrument_data(
            test_data / "19plus_V2_CTD-processing_example_wild_edit.cnv"
        )
        expected_conductivity = expected_dataset.measurements["c0S/m"].values

        dataset = idata.cnv_to_instrument_data(test_data / "19plus_V2_CTD-processing_example.cnv")
        conductivity = dataset.measurements["c0S/m"].values
        flags = dataset.measurements["flag"].values

        wild_edit_output = p.wild_edit(conductivity, flags, 2, 20, 100, 0, False)

        request.node.return_value = wild_edit_output.tolist()
        assert np.all(wild_edit_output == expected_conductivity)


class TestWindowFilter:
    file_prefix = Path(test_data / "19plus_V2_CTD-processing_example")
    cnvdata = idata.cnv_to_instrument_data(f"{file_prefix}.cnv")
    pressure = cnvdata.measurements["prdM"].values
    flags = cnvdata.measurements["flag"].values
    window_width = 5
    half_width = 1  # only applies to gaussian
    offset = 0.25  # only applies to gaussian
    sample_interval = cnvdata.interval_s  # only applies to gaussian

    def test_boxcar_filter(self, request):
        expected_dataset = idata.cnv_to_instrument_data(f"{self.file_prefix}_boxcar_5.cnv")
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.BOXCAR,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            False,
        )

        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

    def test_boxcar_filter_exclude_flags(self, request):
        expected_dataset = idata.cnv_to_instrument_data(
            f"{self.file_prefix}_boxcar_5_excluded.cnv"
        )
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.BOXCAR,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            True,
        )

        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

    def test_cosine_filter(self, request):
        expected_dataset = idata.cnv_to_instrument_data(f"{self.file_prefix}_cosine_5.cnv")
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.COSINE,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            False,
        )

        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

    def test_cosine_filter_exclude_flags(self, request):
        expected_dataset = idata.cnv_to_instrument_data(
            f"{self.file_prefix}_cosine_5_excluded.cnv"
        )
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.COSINE,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            True,
        )

        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

    def test_triangle_filter(self, request):
        expected_dataset = idata.cnv_to_instrument_data(f"{self.file_prefix}_triangle_5.cnv")
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.TRIANGLE,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            False,
        )

        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

    def test_triangle_filter_exclude_flags(self, request):
        expected_dataset = idata.cnv_to_instrument_data(
            f"{self.file_prefix}_triangle_5_excluded.cnv"
        )
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.TRIANGLE,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            True,
        )

        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

    def test_gaussian_filter(self, request):
        expected_dataset = idata.cnv_to_instrument_data(f"{self.file_prefix}_gaussian_5_1_025.cnv")
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.GAUSSIAN,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            False,
        )

        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

    def test_gaussian_filter_exclude_flags(self, request):
        expected_dataset = idata.cnv_to_instrument_data(
            f"{self.file_prefix}_gaussian_5_1_025_excluded.cnv"
        )
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.GAUSSIAN,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            True,
        )

        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

    def test_median_filter(self, request):
        expected_dataset = idata.cnv_to_instrument_data(f"{self.file_prefix}_median_5.cnv")
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.MEDIAN,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            False,
        )
        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

    def test_median_filter_exclude_flags(self, request):
        expected_dataset = idata.cnv_to_instrument_data(
            f"{self.file_prefix}_median_5_excluded.cnv"
        )
        expected_pressure = expected_dataset.measurements["prdM"].values

        filtered_pressure = p.window_filter(
            self.pressure,
            self.flags,
            p.WindowFilterType.MEDIAN,
            self.window_width,
            self.cnvdata.interval_s,
            self.half_width,
            self.offset,
            True,
        )
        request.node.return_value = filtered_pressure.tolist()
        assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)


class TestBuoyancy:
    # fmt: off
    # Testing data comes from a CalCOFI cruise
    temperature = np.asarray(
        [16.7373, 16.5030, 16.1106, 14.3432, 13.0211, 12.0935, 11.3933, 11.2466, 10.9219, 10.4762, 9.9460]
    )
    pressure = np.asarray([10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110])
    salinity = np.asarray(
        [33.2410, 33.2321, 33.2091, 33.1329, 33.0762, 33.1391, 33.2560, 33.4015, 33.5683, 33.6766, 33.7794]
    )
    expected_N2_win30 = np.asarray(
        [-9.990e-29, 5.7702e-05, 1.9197e-04, 2.6735e-04, 2.1888e-04, 2.1620e-04, 1.7374e-04, 1.5761e-04, 1.6905e-04, 1.6099e-04, -9.990e-29]
    )
    expected_N_win30 = np.asarray(
        [-9.990e-29, 4.35, 7.94, 9.37, 8.48, 8.42, 7.55, 7.19, 7.45, 7.27, -9.990e-29]
    )
    expected_E_win30 = np.asarray(
        [-9.990e-29, 5.8901e-06, 1.9596e-05, 2.7290e-05, 2.2342e-05, 2.2069e-05, 1.7735e-05, 1.6089e-05, 1.7256e-05, 1.6433e-05, -9.990e-29]
    )
    expected_E_pow_8_win30 = np.asarray(
        [-9.990e-29, 589.0, 1959.6, 2729.0, 2234.2, 2206.9, 1773.5, 1608.9, 1725.6, 1643.3, -9.990e-29]
    )
    # fmt: on

    def test_buoyancy(self, request):
        output_dataframe = p.buoyancy(
            self.temperature,
            self.salinity,
            self.pressure,
            np.asarray([34.034167]),  # converted from metadata 34.02.03 N in H,M,S
            np.asarray([121.060556]),  # converted from metadata 121 03.38 W in H, M, S
            30,  # window size
            True,
        )

        request.node.return_value = {
            "N2": output_dataframe["N2"].to_list(),
            "N": output_dataframe["N"].to_list(),
            "E": output_dataframe["E"].to_list(),
            "E10^-8": output_dataframe["E10^-8"].to_list(),
        }
        # Comparing EOS-80 to TEOS-10 buoyancy calculations.
        # We do not expect them to agree better than +/-1.5% due to differences in the algorithms
        rel_tol = 0.015  # 1.5%
        assert output_dataframe["N2"].to_numpy() == pytest.approx(
            self.expected_N2_win30, rel=rel_tol
        )
        assert output_dataframe["N"].to_numpy() == pytest.approx(
            self.expected_N_win30, rel=rel_tol
        )
        assert output_dataframe["E"].to_numpy() == pytest.approx(
            self.expected_E_win30, rel=rel_tol
        )
        assert output_dataframe["E10^-8"].to_numpy() == pytest.approx(
            self.expected_E_pow_8_win30, rel=rel_tol
        )

    def test_buoyancy_eos80(self, request):
        output_dataframe = p.buoyancy(
            self.temperature,
            self.salinity,
            self.pressure,
            np.asarray([34.034167]),  # converted from metadata 34.02.03 N in H,M,S
            np.asarray([121.060556]),  # converted from metadata 121 03.38 W in H, M, S
            30,  # window size
            False,
        )

        request.node.return_value = {
            "N2": output_dataframe["N2"].to_list(),
            "N": output_dataframe["N"].to_list(),
            "E": output_dataframe["E"].to_list(),
            "E10^-8": output_dataframe["E10^-8"].to_list(),
        }
        # Comparing SBE Data Processing C++ to local Python results using the same EOS-80 calculations.
        # We expect very very close agreement: << 1% differnce
        rel_tol = 0.0026  # 0.26%
        assert output_dataframe["N2"].to_numpy() == pytest.approx(
            self.expected_N2_win30, rel=rel_tol
        )
        assert output_dataframe["N"].to_numpy() == pytest.approx(
            self.expected_N_win30, rel=rel_tol
        )
        assert output_dataframe["E"].to_numpy() == pytest.approx(
            self.expected_E_win30, rel=rel_tol
        )
        assert output_dataframe["E10^-8"].to_numpy() == pytest.approx(
            self.expected_E_pow_8_win30, rel=rel_tol
        )


class TestNitrate:
    dac_min = -5
    dac_max = 100
    voltages = np.array([0.095, 1, 2, 3, 4.095])

    def test_convert_nitrate_umno3(self, request):
        expected_umno3 = np.array([-5, 18.75625, 45.00625, 71.25625, 100])
        nitrate = c.convert_nitrate(self.voltages, dac_min=self.dac_min, dac_max=self.dac_max)
        request.node.return_value = nitrate.tolist()
        assert np.allclose(expected_umno3, nitrate, atol=0.000001)

    def test_convert_nitrate_mgnl(self, request):
        expected_mgnl = np.array([-0.070035, 0.26271879375, 0.63040254375, 0.99808629375, 1.4007])
        nitrate = c.convert_nitrate(
            self.voltages, dac_min=self.dac_min, dac_max=self.dac_max, units="mgNL"
        )
        request.node.return_value = nitrate.tolist()
        assert np.allclose(expected_mgnl, nitrate, atol=0.000001)


class TestSplit:
    def test_split(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped.cnv"
        source_down = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_downcast.cnv"
        source_up = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_upcast.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected_down = idata.cnv_to_instrument_data(source_down)._to_dataframe()
        expected_up = idata.cnv_to_instrument_data(source_up)._to_dataframe()

        down, up = p.split(data, "prdM")

        for variable in expected_down.columns:
            tolerance = get_tolerance(expected_down[variable])
            assert np.allclose(expected_down[variable], down[variable], rtol=0, atol=10**tolerance)
            assert np.allclose(expected_up[variable], up[variable], rtol=0, atol=10**tolerance)

    def test_split_with_flags(self, request):
        source_in = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit.cnv"
        source_down = (
            test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_downcast.cnv"
        )
        source_up = test_data / "SBE19plus_01906398_2019_07_15_0033_cropped_loopedit_upcast.cnv"
        data = idata.cnv_to_instrument_data(source_in)._to_dataframe()
        expected_down = idata.cnv_to_instrument_data(source_down)._to_dataframe()
        expected_up = idata.cnv_to_instrument_data(source_up)._to_dataframe()

        down, up = p.split(data, "prdM", exclude_bad_scans=True)

        for variable in expected_down.columns:
            tolerance = get_tolerance(expected_down[variable])
            assert np.allclose(expected_down[variable], down[variable], rtol=0, atol=10**tolerance)
            assert np.allclose(expected_up[variable], up[variable], rtol=0, atol=10**tolerance)



"""Data conversion unit tests."""

# Native imports

# Third-party imports
from pathlib import Path
import numpy as np
import pandas as pd
import pytest

# Sea-Bird imports

# Internal imports
import seabirdscientific.conversion as dc
import seabirdscientific.instrument_data as id

import test_coefficients as tc

test_data = Path("./tests/resources/test-data")


class TestConvertTemperature:
    # test temperature raw values
    test_temp_vals = np.array([322798, 322808, 322827, 322838])
    test_temp_no_mv_r = np.array([2864.51635696, 2864.61073548, 2864.79005751, 2864.89387722])

    def test_convert_temperature_90C(self, request):
        expected = [20.4459, 20.4451, 20.4436, 20.4427]
        result = dc.convert_temperature(
            self.test_temp_vals,
            tc.temperature_coefs_sn6130,
            "ITS90",
            "C",
            True,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_90F(self, request):
        expected = [68.8027, 68.8012, 68.7984, 68.7968]
        result = dc.convert_temperature(
            self.test_temp_vals,
            tc.temperature_coefs_sn6130,
            "ITS90",
            "F",
            True,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_68C(self, request):
        expected = [20.4508, 20.4500, 20.4485, 20.4476]
        result = dc.convert_temperature(
            self.test_temp_vals,
            tc.temperature_coefs_sn6130,
            "IPTS68",
            "C",
            True,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_68F(self, request):
        expected = [68.8115, 68.8100, 68.8073, 68.8057]
        result = dc.convert_temperature(
            self.test_temp_vals,
            tc.temperature_coefs_sn6130,
            "IPTS68",
            "F",
            True,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_90C_no_mv_r(self, request):
        expected = [20.4459, 20.4451, 20.4436, 20.4427]
        result = dc.convert_temperature(
            self.test_temp_no_mv_r,
            tc.temperature_coefs_sn6130,
            "ITS90",
            "C",
            False,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)


class TestConvertTemperatureFrequency:
    test_temp_freq_vals = np.array(
        [4651.168, 4650.961, 4650.219, 4649.523, 4649.418, 4649.879, 3771.953, 4084.508]
    )

    def test_convert_temperature_frequency_90C(self, request):
        expected = [16.6799, 16.6777, 16.6698, 16.6624, 16.6613, 16.6662, 6.6247, 10.3732]
        result = dc.convert_temperature_frequency(
            self.test_temp_freq_vals,
            tc.temperature_frequency_coefs_sn5102,
            "ITS90",
            "C",
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_frequency_90F(self, request):
        expected = [62.0238, 62.0199, 62.0056, 61.9923, 61.9903, 61.9991, 43.9244, 50.6718]
        result = dc.convert_temperature_frequency(
            self.test_temp_freq_vals,
            tc.temperature_frequency_coefs_sn5102,
            "ITS90",
            "F",
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_frequency_68C(self, request):
        expected = [16.6839, 16.6817, 16.6738, 16.6664, 16.6653, 16.6702, 6.6263, 10.3757]
        result = dc.convert_temperature_frequency(
            self.test_temp_freq_vals,
            tc.temperature_frequency_coefs_sn5102,
            "IPTS68",
            "C",
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_frequency_68F(self, request):
        expected = [62.0310, 62.0271, 62.0128, 61.9995, 61.9975, 62.0063, 43.9273, 50.6763]
        result = dc.convert_temperature_frequency(
            self.test_temp_freq_vals,
            tc.temperature_frequency_coefs_sn5102,
            "IPTS68",
            "F",
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)


class TestConvertPressure:
    # test pressure raw values
    test_pressure_vals = np.array([533539, 533538, 533540, 533537])

    # test pressure temperature compensation raw values
    test_compensation_vals = np.array([20625, 20626, 20626, 20626]) / 13107

    def test_convert_pressure_psia(self, request):
        expected = [14.547, 14.546, 14.549, 14.544]
        result = dc.convert_pressure(
            self.test_pressure_vals,
            self.test_compensation_vals,
            tc.pressure_coefs_sn6130,
            "psia",
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-3)

    def test_convert_pressure_dbar(self, request):
        expected = [-0.105, -0.106, -0.104, -0.107]
        result = dc.convert_pressure(
            self.test_pressure_vals,
            self.test_compensation_vals,
            tc.pressure_coefs_sn6130,
            "dbar",
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-3)


class TestDigiquartzPressure:
    # test pressure raw values
    test_pressure_vals = np.array(
        [33302.258, 33302.250, 33302.230, 33302.285, 33302.250, 33302.258]
    )

    # test pressure temperature compensation raw values
    test_compensation_vals = np.array([2498.0, 2499.0, 2498.0, 2499.0, 2499.0, 2498.0])

    def test_convert_pressure_digiquartz_dbar(self, request):
        expected = [2.099, 2.085, 2.051, 2.146, 2.085, 2.099]
        result = dc.convert_pressure_digiquartz(
            self.test_pressure_vals,
            self.test_compensation_vals,
            tc.pressure_digiquartz_coefs_sn5102,
            "dbar",
            0.25,
        )
        # apply slope and offset values
        result = result * 1.00001297 - 2.46118
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-3)


class TestConductivity19plus:
    cnv_path = test_data / "19plus_V2_CTD-processing_example.cnv"
    hex_path = test_data / "19plus_V2.hex"

    def test_convert_conductivity(self, request):
        # expected_data = id.cnv_to_instrument_data(self.cnv_path)

        raw = id.read_hex_file(
            self.hex_path,
            id.InstrumentType.SBE19Plus,
            [
                id.Sensors.Temperature,
                id.Sensors.Conductivity,
                id.Sensors.Pressure,
                id.Sensors.ExtVolt0,
                id.Sensors.ExtVolt1,
                id.Sensors.ExtVolt2,
                id.Sensors.ExtVolt4,
            ],
        )

        # test_cond_data = np.array([675415, 675404, 675405, 675398, 675391, 675379])

        # test_temp_data = np.array([20.4459, 20.4451, 20.4436, 20.4427, 20.4413, 20.4401])
        temperature = dc.convert_temperature(
            raw["temperature"][0:6].values,
            tc.temperature_coefs_sn6130,
            "IPTS68",
            "C",
            True,
        )

        # test_press_data = np.array([-0.105, -0.106, -0.104, -0.107, -0.105, -0.104])
        pressure = dc.convert_pressure(
            raw["pressure"][0:6].values,
            raw["temperature compensation"][0:6].values,
            tc.pressure_coefs_sn6130,
            "psia",
        )
        expected = [0.008453, 0.008420, 0.008423, 0.008402, 0.008380, 0.008344]
        result = dc.convert_conductivity(
            raw["conductivity"][0:6].values,
            temperature,
            pressure,
            tc.conductivity_coefs_sn6130,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-6)


class TestConductivity37SM:
    cnv_path = test_data / "SBE37SM-RS232_03716125_2017_11_16.cnv"
    hex_path = test_data / "SBE37SM-RS232_03716125_2017_11_16.hex"

    def test_convert_conductivity(self, request):
        # expected_data = id.cnv_to_instrument_data(self.cnv_path)

        raw = id.read_hex_file(
            self.hex_path,
            id.InstrumentType.SBE37SM,
            [
                id.Sensors.Temperature,
                id.Sensors.Conductivity,
                id.Sensors.Pressure,
            ],
        )

        temperature = dc.convert_temperature(
            raw["temperature"][0:6],
            tc.temperature_coefs_sn6130,
            "IPTS68",
            "C",
            True,
        )

        # test_press_data = np.array([-0.105, -0.106, -0.104, -0.107, -0.105, -0.104])
        pressure = dc.convert_pressure(
            raw["pressure"][0:6],
            raw["temperature compensation"][0:6],
            tc.pressure_coefs_sn16125,
            "psia",
        )
        expected = [2.711842, 2.715786, 2.715857, 2.715846, 2.715846, 2.715857]
        result = dc.convert_conductivity(
            raw["conductivity"][0:6],
            temperature,
            pressure,
            tc.conductivity_coefs_sn16125,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-4)


class TestDeriveDensity:
    data_path = test_data / "SBE37SM-derived.asc"
    data = pd.read_csv(data_path)

    @pytest.mark.parametrize(
        "reference_pressure, expected_column",
        [
            (0.0, "gsw_sigma0A0"),
            (1000.0, "gsw_sigma1A0"),
            (2000.0, "gsw_sigma2A0"),
            (3000.0, "gsw_sigma3A0"),
            (4000.0, "gsw_sigma4A0"),
        ],
    )
    def test_derive_potential_density_from_t_s_p_pass(
        self, request, reference_pressure, expected_column
    ):
        temperature = self.data["t090C"].values
        salinity = self.data["sal00"].values
        pressure = self.data["prM"].values
        expected = self.data[expected_column].values
        result = dc.potential_density_from_t_s_p(
            temperature,
            salinity,
            pressure,
            reference_pressure=reference_pressure,
        )
        request.node.return_value = result.tolist()
        # TODO: improve passing condition for all instances of allclose
        assert np.allclose(result, expected, atol=0.1)

    def test_derive_density_from_t_s_p_pass(self, request):
        temperature = self.data["t090C"].values
        salinity = self.data["sal00"].values
        pressure = self.data["prM"].values
        expected = self.data["gsw_densityA0"].values
        result = dc.density_from_t_s_p(temperature, salinity, pressure)

        request.node.return_value = result.tolist()
        assert np.allclose(result, expected, atol=0.1)

    @pytest.mark.parametrize(
        "reference_pressure, expected_column",
        [
            (0.0, "gsw_sigma0A0"),
            (1000.0, "gsw_sigma1A0"),
            (2000.0, "gsw_sigma2A0"),
            (3000.0, "gsw_sigma3A0"),
            (4000.0, "gsw_sigma4A0"),
        ],
    )
    def test_derive_potential_density_from_t_c_p_pass(
        self, request, reference_pressure, expected_column
    ):
        temperature = self.data["t090C"].values
        conductivity = self.data["c0S/m"].values * 10.0
        pressure = self.data["prM"].values
        expected = self.data[expected_column].values
        result = dc.potential_density_from_t_c_p(
            temperature,
            conductivity,
            pressure,
            reference_pressure=reference_pressure,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(result, expected, atol=0.1)

    def test_derive_density_from_t_c_p_pass(self, request):
        temperature = self.data["t090C"].values
        conductivity = self.data["c0S/m"].values * 10.0
        pressure = self.data["prM"].values
        expected = self.data["gsw_densityA0"].values
        result = dc.density_from_t_c_p(temperature, conductivity, pressure)

        request.node.return_value = result.tolist()
        assert np.allclose(result, expected, atol=0.1)


class TestDepthFromPressure:
    data_path = test_data / "SBE37SM.asc"
    data = pd.read_csv(data_path).loc[6:10]

    @pytest.mark.parametrize(
        "pressure, pressure_units, depth, depth_units",
        [
            ("prdM", "dbar", "depSM", "m"),
            ("prdM", "dbar", "depSF", "ft"),
            ("prdE", "psi", "depSM", "m"),
            ("prdE", "psi", "depSF", "ft"),
        ],
    )
    def test_depth_from_pressure_pass(self, request, pressure, pressure_units, depth, depth_units):
        expected_depth = self.data[depth].values
        pressure = self.data[pressure].values
        result_depth = dc.depth_from_pressure(pressure, 0, depth_units, pressure_units)
        request.node.return_value = result_depth.tolist()
        assert np.allclose(expected_depth, result_depth, atol=0.002)


class TestConvertOxygen:
    def test_convert_sbe63_oxygen(self, request):
        raw_oxygen = np.array([31.06, 31.66, 32.59, 33.92, 34.82, 35.44])
        pressure = np.array([0, 0, 0, 0, 0, 0])
        raw_temperature = np.array([0.6, 0.5, 0.4, 0.35, 0.3, 0.25])
        salinity = np.array([0, 0, 0, 0, 0, 0])
        expected = np.array([0.93, 0.688, 0.459, 0.304, 0.206, 0.137])

        result = dc.convert_sbe63_oxygen(
            raw_oxygen,
            raw_temperature,
            pressure,
            salinity,
            tc.oxygen_63_coefs_sn2568,
            tc.thermistor_63_coefs_sn2568,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-3)

    def test_convert_sbe63_oxygen_from_hex(self, request):
        raw_oxygen = np.array([16.774, 16.775, 16.779, 16.778, 16.774, 16.779])
        pressure = np.array([-0.057, -0.062, -0.057, -0.056, -0.068, -0.056])
        raw_temperature = np.array([0.581763, 0.581758, 0.581753, 0.581737, 0.581725, 0.581717])
        salinity = np.array([0.0115, 0.0115, 0.0115, 0.0115, 0.0115, 0.0115])
        expected = np.array([5.872, 5.872, 5.868, 5.869, 5.872, 5.868])

        result = dc.convert_sbe63_oxygen(
            raw_oxygen,
            raw_temperature,
            pressure,
            salinity,
            tc.oxygen_63_coefs_sn11459,
            tc.thermistor_63_coefs_sn11459,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-3)

    def test_convert_sbe43_oxygen(self, request):
        # From O3287.pdf in the shared calibration folder
        raw_oxygen = np.array([0.725, 0.756, 0.803, 0.874, 0.925, 0.96, 1.332, 1.435, 1.595, 1.81])
        pressure = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        temperature = np.array([2, 6, 12, 20, 26, 30, 2, 6, 12, 20])
        salinity = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        expected = np.array([1.11, 1.12, 1.13, 1.16, 1.18, 1.18, 3.9, 3.9, 3.92, 3.95])

        result = dc._convert_sbe43_oxygen(
            raw_oxygen,
            temperature,
            pressure,
            salinity,
            tc.oxygen_43_coefs_sn3287,
            0,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-2)

    def test_convert_sbe43_oxygen_from_hex(self, request):
        # fmt: off
        # From SBE19plus_01906398_2019_07_15_0033.hex
        raw_oxygen = np.array(
            [2.5575, 2.5586, 2.5606, 2.5627, 2.5638, 2.5637, 2.5635, 2.5629, 2.5621, 2.5618]
        )
        pressure = np.array(
            [-0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.011, 0.107, 0.351]
        )
        temperature = np.array(
            [25.3427, 25.3408, 25.3387, 25.3363, 25.3341, 25.3326, 25.3316, 25.3302, 25.3377, 25.5433]
        )
        salinity = np.array(
            [0.4373, 0.5592, 0.5865, 0.5095, 0.4621, 0.4119, 0.3936, 0.3463, 4.9297, 6.5098]
        )
        expected = np.array(
            [4.4728, 4.4722, 4.4762, 4.4828, 4.4867, 4.4879, 4.488, 4.488, 4.3707, 4.3148]
        )
        # fmt: on
        result = dc._convert_sbe43_oxygen(
            raw_oxygen,
            temperature,
            pressure,
            salinity,
            tc.oxygen_43_coefs_sn1686,
            0,
        )
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-3)

    def test_convert_sbe43_oxygen_from_hex_with_hysteresis(self, request):
        # fmt: off
        # From SBE19plus_01906398_2019_07_15_0033.hex
        # TODO: hysteresis correction only has a real impact on deep data, will need some to better validate this
        raw_oxygen = np.array(
            [2.5575, 2.5586, 2.5606, 2.5627, 2.5638, 2.5637, 2.5635, 2.5629, 2.5621, 2.5618]
        )
        pressure = np.array(
            [-0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.011, 0.107, 0.351]
        )
        temperature = np.array(
            [25.3427, 25.3408, 25.3387, 25.3363, 25.3341, 25.3326, 25.3316, 25.3302, 25.3377, 25.5433]
        )
        salinity = np.array(
            [0.4373, 0.5592, 0.5865, 0.5095, 0.4621, 0.4119, 0.3936, 0.3463, 4.9297, 6.5098]
        )
        expected = np.array(
            [4.4728, 4.4722, 4.4762, 4.4828, 4.4867, 4.4879, 4.488, 4.488, 4.3707, 4.3148]
        )
        # fmt: on
        result = dc.convert_sbe43_oxygen(
            raw_oxygen,
            temperature,
            pressure,
            salinity,
            tc.oxygen_43_coefs_sn1686,
            False,
            True,
            1,
            0.25,
        )

        request.node.return_value = result.tolist()
        assert np.allclose(result, expected, rtol=0, atol=1e-3)

    def test_convert_sbe43_oxygen_from_hex_with_tau_correction(self, request):
        # fmt: off
        # From SBE19plus_01906398_2019_07_15_0033.hex
        raw_oxygen = np.asarray(
            [2.5575, 2.5586, 2.5606, 2.5627, 2.5638, 2.5637, 2.5635, 2.5629, 2.5621, 2.5618]
        )
        pressure = np.asarray(
            [-0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.011, 0.107, 0.351]
        )
        temperature = np.asarray(
            [25.3427, 25.3408, 25.3387, 25.3363, 25.3341, 25.3326, 25.3316, 25.3302, 25.3377, 25.5433]
        )
        salinity = np.asarray(
            [0.4373, 0.5592, 0.5865, 0.5095, 0.4621, 0.4119, 0.3936, 0.3463, 4.9297, 6.5098]
        )
        expected = [4.4729, 4.4723, 4.4884, 4.4927, 4.4916, 4.4879, 4.4849, 4.4841, 4.3707, 4.3148]
        # fmt: on
        result = dc.convert_sbe43_oxygen(
            raw_oxygen,
            temperature,
            pressure,
            salinity,
            tc.oxygen_43_coefs_sn1686,
            True,
            False,
            1,
            0.25,
        )

        request.node.return_value = result.tolist()
        assert np.allclose(result, expected, rtol=0, atol=1e-4)

    def test_convert_to_mg_per_l(self, request):
        oxMlPerL = np.array(
            [4.4728, 4.4722, 4.4762, 4.4828, 4.4867, 4.4879, 4.488, 4.488, 4.3707, 4.3148]
        )
        expected = [6.3921, 6.3913, 6.3969, 6.4064, 6.4119, 6.4137, 6.4138, 6.4138, 6.2461, 6.1663]
        result = dc.convert_oxygen_to_mg_per_l(oxMlPerL)
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-3)

    def test_convert_to_umol_per_kg(self, request):
        # fmt: off
        oxMlPerL = np.array(
            [4.4728, 4.4722, 4.4762, 4.4828, 4.4867, 4.4879, 4.488, 4.488, 4.3707, 4.3148]
        )
        expected = np.array(
            [200.3, 200.254, 200.427, 200.735, 200.916, 200.979, 200.984, 200.991, 195.064, 192.356]
        )
        potentialDensity = np.array(
            [-2.7113, -2.6188, -2.5977, -2.6552, -2.6903, -2.7279, -2.7414, -2.7768, 0.6655, 1.7939]
        )
        # fmt: on
        result = dc.convert_oxygen_to_umol_per_kg(oxMlPerL, potentialDensity)
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-2)


class TestConvertChlorophylla:
    def test_convert_eco(self, request):
        # fmt: off
        rawAnalog = np.array(
            [0.0949, 0.0948, 0.0960, 0.0961, 0.0962, 0.0959, 0.1013, 0.1012, 0.1015, 0.1012, 0.1003, 0.0999, 0.0999, 0.0996]
        )
        expected = np.array(
            [0.2691, 0.2683, 0.2798, 0.2813, 0.2821, 0.279, 0.3332, 0.3317, 0.3355, 0.3317, 0.3233, 0.3187, 0.3195, 0.3157]
        )
        # fmt: on
        result = dc.convert_eco(rawAnalog, tc.chlorophyll_a_coefs_sn6130)
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-2)


class TestConvertpH:
    def test_convert_sbe18_ph(self, request):
        # fmt: off
        rawVolts = np.array(
            [2.9507, 2.9518, 2.9522, 2.9517, 2.9521, 2.9519, 2.9517, 2.9526, 2.9525, 2.9537, 2.953, 2.9527]
        )
        temperatureC = np.array(
            [26.6927, 26.6994, 26.6624, 26.6122, 26.5808, 26.5621, 26.5503, 26.5505, 26.569, 26.5954, 26.6182, 26.6359]
        )
        expected = np.array(
            [8.587, 8.591, 8.593, 8.591, 8.593, 8.592, 8.592, 8.595, 8.594, 8.599, 8.596, 8.595]
        )
        # fmt: on
        result = dc.convert_sbe18_ph(rawVolts, temperatureC, tc.ph_coefs_sn0762)
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-3)


class TestPARlogarithmic:
    def test_convert_par_logarithmic(self, request):
        # fmt: off
        rawVolts = np.array(
            [1.176241702, 1.175249866, 1.174944686, 1.176394293, 0.089951934, 0.090562294, 0.090638590, 0.091172655, 0.090485999, 0.090943770, 0.090409704, 0.091096360, 0.093537804]
        )
        expected = np.array(
            [0.81605, 0.81394, 0.81330, 0.81637, 0.04817, 0.04824, 0.04825, 0.04832, 0.04823, 0.04829, 0.04822, 0.04831, 0.04862]
        )
        # fmt :on
        result = dc.convert_par_logarithmic(rawVolts, tc.par_coefs_sn0411)
        request.node.return_value = result.tolist()
        assert np.allclose(expected, result, rtol=0, atol=1e-3)


# class TestContourFromTSP:
#     # Note: this class doesn't actually test anything and is only for debug
#     data_path = test_data / 'SBE37SM-RS232_03711722_2015_11_18_subset_derived.asc'
#     data = pd.read_csv(data_path)

#     def test_contour_from_t_s_p_pass(self, request):
#         temperature = self.data['t090C'].values
#         salinity = self.data['sal00'].values
#         pressure = self.data['prM'].values
#         contour_data = c.contour_from_t_s_p(
#             temperature, salinity, pressure
#         )
#         assert(True)


class TestSeaFETPH:
    internal_ph_counts = np.array([5105334, 5105384, 5105350, 5105505, 5105347])
    external_ph_counts = np.array([5139133, 5139214, 5139120, 5139294, 5139183])
    expected_internal_ph_volts = np.array([-0.978492, -0.978477, -0.978487, -0.978441, -0.978488])
    expected_external_ph_volts = np.array([-0.968419, -0.968395, -0.968423, -0.968371, -0.968404])
    expected_internal_ph = np.array([6.8897, 6.89, 6.8898, 6.8906, 6.8898])
    expected_external_ph = np.array([6.2466, 6.247, 6.2466, 6.2474, 6.2469])
    ph_temperature = np.array([22.751, 22.7514, 22.752, 22.752, 22.752])

    def ph_voltage_to_counts(self, voltage: np.ndarray):
        adc_vref = 2.5
        gain = 1
        adc_23bit = 8388608
        ph_counts = adc_23bit * (voltage * gain / adc_vref + 1)
        return ph_counts

    def test_convert_ph_voltage_counts(self, request):
        internal_ph_volts = dc.convert_ph_voltage_counts(self.internal_ph_counts)
        request.node.return_value = internal_ph_volts.tolist()
        assert np.allclose(internal_ph_volts, self.expected_internal_ph_volts, atol=1e-6)

    def test_convert_internal_seafet_ph(self, request):
        internal_ph = dc.convert_internal_seafet_ph(
            raw_ph=self.internal_ph_counts,
            temperature=self.ph_temperature,
            coefs=tc.ph_seafet_internal_coefs,
        )
        request.node.return_value = internal_ph.tolist()
        assert np.allclose(internal_ph, self.expected_internal_ph, atol=1e-6)

    def test_convert_external_seafet_ph(self, request):
        external_ph = dc.convert_external_seafet_ph(
            ph_counts=self.external_ph_counts,
            temperature=self.ph_temperature,
            salinity=35,
            pressure=0,
            coefs=tc.ph_seafet_external_coefs,
        )
        request.node.return_value = external_ph.tolist()
        assert np.allclose(external_ph, self.expected_external_ph, atol=1e-6)

    def test_internal_shallow_ph_app_note_99(self, request):
        # from application note 99
        ph_counts = self.ph_voltage_to_counts(-1.010404)
        # coefs = cc.PHSeaFETInternalCoefficients(k0=-1.438788, k2=-1.304895e-3)
        internal_ph = dc.convert_internal_seafet_ph(
            raw_ph=ph_counts,
            temperature=15.8735,
            coefs=tc.internal_ph_coefs,
        )
        assert np.allclose(internal_ph, 7.8310, atol=1e-4)

    def test_external_shallow_seafet_seaphox_ph_app_note_99(self, request):
        # from application note 99
        # Shallow pH is effectively the same as deep pH with pressure
        # set to 0. Since the deep pH example in app note 99 is valid,
        # there must be a typo in the shellow pH example. This test uses
        # the same inputs as the shallow example and fixes the result
        external_ph = dc.convert_external_seafet_ph(
            raw_ph=-0.965858,
            temperature=15.8735,
            salinity=36.817,
            coefs=tc.external_shallow_ph_coefs,
            ph_units="volts",
        )
        assert np.allclose(external_ph, 7.9224, atol=1e-4)

    def test_external_deep_seaphox_float_ph_app_note_99(self, request):
        # from application note 99
        external_ph = dc.convert_external_seafet_ph(
            raw_ph=-0.885081,
            temperature=23.4169,
            salinity=34.812,
            pressure=100,
            coefs=tc.external_ph_coefs,
            ph_units="volts",
        )
        assert np.allclose(external_ph, 7.9394, atol=1e-4)

    def test_convert_external_seafet_ph_legacy(self, request):
        # from application note 99
        external_ph = dc.convert_external_seafet_ph(
            ph_counts=self.ph_voltage_to_counts(-0.885081),
            temperature=23.4169,
            salinity=34.812,
            pressure=100,
            coefs=tc.external_ph_coefs,
            formula_version="legacy",
        )
        assert np.allclose(external_ph, 7.939415, atol=1e-6)

    def test_convert_external_seafet_ph_legacy_volts(self, request):
        # from application note 99
        external_ph = dc.convert_external_seafet_ph(
            raw_ph=-0.885081,
            temperature=23.4169,
            salinity=34.812,
            pressure=100,
            coefs=tc.external_ph_coefs,
            ph_units="volts",
            formula_version="legacy",
        )
        assert np.allclose(external_ph, 7.939415, atol=1e-6)

    def test_convert_external_seafet_ph_v1p3(self, request):
        # from application note 99
        external_ph = dc.convert_external_seafet_ph(
            raw_ph=-0.885081,
            temperature=23.4169,
            salinity=34.812,
            pressure=100,
            coefs=tc.external_ph_coefs,
            ph_units="volts",
        )
        assert np.allclose(external_ph, 7.939406, atol=1e-6)


    def test_convert_external_seafet_ph_v1p3_shallow(self, request):
        # from application note 99
        external_ph = dc.convert_external_seafet_ph(
            raw_ph=-0.900081,
            temperature=23.4169,
            salinity=34.812,
            pressure=0,
            coefs=tc.external_ph_coefs,
            ph_units="volts",
        )
        assert np.allclose(external_ph, 7.6653, atol=1e-4)

    def test_convert_external_seafet_ph_argo_example(self, request):
        # from Processing BGC-Argo pH data at the DAC level
        # https://archimer.ifremer.fr/doc/00460/57195/
        expected_ph = [7.9870, 7.9788, 7.9105, 7.7846, 7.7326, 7.7307]
        external_ph = dc.convert_external_seafet_ph(
            raw_ph=np.array([-0.82966, -0.83049, -0.83921, -0.85150, -0.85968, -0.86191]),
            temperature=np.array([29.545, 29.607, 25.741, 21.094, 15.725, 13.407]),
            salinity=np.array([34.637, 34.767, 35.104, 35.080, 34.886, 34.890]),
            pressure=np.array([2.2, 42, 82, 122, 162, 202]),
            coefs=tc.external_ph_k2_poly_coefs,
            ph_units="volts",
            formula_version="1.3",
        )
        assert np.allclose(external_ph, expected_ph, atol=1e-4)


class TestInternalSeaFETTemperature:
    def test_convert_internal_seafet_temperature(self, request):
        temperature_counts = np.array([25616, 25600])
        expected_temperature = np.array([21.8335, 21.7906])
        temperature = dc.convert_internal_seafet_temperature(temperature_counts=temperature_counts)
        request.node.return_value = temperature.tolist()
        assert np.allclose(expected_temperature, temperature, atol=1e-6)


class TestSeaFETRelativeHumidity:
    def test_convert_seafet_relative_humidity(self, request):
        humidity_counts = np.array([24096, 24160])
        temperature_counts = np.array([25616, 25600])
        expected_humidity = np.array([39.4845, 39.6001])

        temperature = dc.convert_internal_seafet_temperature(temperature_counts=temperature_counts)
        humidity = dc.convert_seafet_relative_humidity(
            humidity_counts=humidity_counts, temperature=temperature
        )

        request.node.return_value = humidity.tolist()
        assert np.allclose(expected_humidity, humidity, atol=1e-6)


class TestConvertSBE63Oxygen:
    raw_oxygen = np.array([31.06, 31.66, 32.59, 33.92, 34.82, 35.44, 35.44, 35.44])
    pressure = np.array([0, 0, 0, 0, 0, 0, 1000, 100])
    temperature = np.array([30, 26, 20, 12, 6, 2, 2, 2])
    salinity = np.array([0, 0, 0, 0, 0, 0, 0, 35])  #  salinity is 0 PSU during calibration
    expected_oxygen = np.array([0.706, 0.74, 0.799, 0.892, 1.005, 1.095, 1.1398, 0.8647])

    def test_convert_sbe63_oxygen(self, request):
        oxygen = dc.convert_sbe63_oxygen(
            self.raw_oxygen,
            self.temperature,
            self.pressure,
            self.salinity,
            tc.oxygen_63_coefs_sn2568,
            tc.thermistor_63_coefs_sn2568,
            "C",
        )
        # TODO: fix tolerance in TKIT-110
        request.node.return_value = oxygen.tolist()
        assert np.allclose(self.expected_oxygen, oxygen, atol=1e-3)

    def test_thermistor_temperature(self, request):
        raw_temperature = np.array([1.12015, 1.12015, 1.12016, 1.12016, 0.99562, 0.82934, 0.64528])

        expected = np.array([2.0002, 2.0002, 1.9999, 1.9999, 5.9998, 12.0, 19.9998])
        thermistor_temperature = dc.convert_sbe63_thermistor(
            raw_temperature, tc.thermistor_63_coefs_sn2568
        )

        request.node.return_value = thermistor_temperature.tolist()
        assert np.allclose(expected, thermistor_temperature, atol=1e-6)


class TestSPAR:
    def test_surface_par(self, request):
        spar_raw = np.array([0.03296703296703297])
        expected_spar_biospherical = np.array([51.41538])  # from SBE data processing
        expected_spar_linear = np.array([-6636])  # from SBE data processing
        expected_spar_logarithmic = np.array([1127.0])  # from SBE data processing

        spar_biospherical = dc.convert_spar_biospherical(spar_raw, tc.spar_coefs)
        spar_linear = dc.convert_spar_linear(spar_raw, tc.spar_coefs)
        spar_logarithmic = dc.convert_spar_logarithmic(spar_raw, tc.spar_coefs)

        assert np.allclose(expected_spar_biospherical, spar_biospherical, atol=1e-5)
        assert np.allclose(expected_spar_linear, spar_linear, atol=0)
        assert np.allclose(expected_spar_logarithmic, spar_logarithmic, atol=1e-1)


class TestAltimeter:
    def test_altimeter(self, request):
        altimeter_raw = np.array([3.1893, 1.1807, 2.3797])
        expected = np.array([63.79, 23.61, 47.59])  ## meters, from SBE data processing

        height = dc.convert_altimeter(altimeter_raw, tc.altimeter_coefs)

        assert np.allclose(expected, height, atol=1e-2)



"""A collection of classes and functions related to the processing of
instrument data.
"""

# Classes:
#   InstrumentType (Enum)
#   HexDataTypes (Enum)
#   Sensors (Enum)
#   MeasurementSeries
#   InstrumentData
# Functions:
#   cnv_to_instrument_data (Path) -> InstrumentData
#   fix_exponents (List[str]) -> List[str]
#   read_hex_file (str, InstrumentType, List[Sensors], bool) -> pd.DataFrame
#   preallocate_dataframe (InstrumentType, str, List[Sensors], bool, int) -> pd.DataFrame
#   read_hex (InstrumentType, str, List[Sensors], bool) -> dict
#   read_SBE19plus_format_0
#   read_SBE37SM_format_0
#   read_SBE39plus_format_0
#   read_seafet_format_0
#   read_SBE911plus_format_0

# Native imports
import builtins
from enum import Enum
from dataclasses import dataclass
from datetime import date, datetime, timedelta
from logging import getLogger
from typing import List, Dict, Optional, Union
from pathlib import Path
import warnings

# Third-party imports
import numpy as np
import pandas as pd

# Sea-Bird imports

# Internal imports

logger = getLogger(__name__)

COUNTS_TO_VOLTS = 13107
SECONDS_BETWEEN_EPOCH_AND_2000 = 946684800


class InstrumentType(Enum):
    """The type of instrument that generated the hex file being read"""

    SBE37SM = "37-SM"
    SBE37SMP = "37-SMP"
    SBE37SMPODO = "37-SMP-ODO"
    SBE37SI = "37-SI"
    SBE37SIP = "37-SIP"
    SBE37IM = "37-IM"
    SBE37IMP = "37-IMP"
    SBE37IMPODO = "37-IMP-ODO"
    SBE19Plus = "19plus"  # change enums to UPPER_CASE for TKIT-75
    SBE16Plus = "16plus"
    SBE39Plus = "39plus"
    SBE39PlusIM = "39plus-IM"
    SBE911Plus = "911plus"
    SeaFET2 = "SeaFET2"
    SeapHox2 = "SeapHox2"
    HydroCAT = "HydroCAT"
    HydroCATODO = "HydroCAT-ODO"


class HexDataTypes(Enum):
    """Possible data types in hex files"""

    temperature = (  # change enums to UPPER_CASE for TKIT-75
        "temperature"
    )
    secondaryTemperature = (  # change enums to UPPER_CASE for TKIT-75
        "secondary temperature"
    )
    conductivity = (  # change enums to UPPER_CASE for TKIT-75
        "conductivity"
    )
    secondaryConductivity = (  # change enums to UPPER_CASE for TKIT-75
        "secondary conductivity"
    )

    pressure = "pressure"  # change enums to UPPER_CASE for TKIT-75
    digiquartzPressure = "digiquartz pressure"  # change enums to UPPER_CASE for TKIT-75
    temperatureCompensation = "temperature compensation"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt0 = "volt 0"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt1 = "volt 1"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt2 = "volt 2"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt3 = "volt 3"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt4 = "volt 4"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt5 = "volt 5"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt6 = "volt 6"
    ExtVolt7 = "volt 7"
    SBE38temperature = "SBE38 temperature"  # change enums to UPPER_CASE for TKIT-75
    wetlabs0 = "wetlabs - channel 0"  # change enums to UPPER_CASE for TKIT-75
    wetlabs1 = "wetlabs - channel 1"  # change enums to UPPER_CASE for TKIT-75
    wetlabs2 = "wetlabs - channel 2"  # change enums to UPPER_CASE for TKIT-75
    GTDpressure = (  # change enums to UPPER_CASE for TKIT-75
        "GTD pressure"
    )
    GTDtemperature = (  # change enums to UPPER_CASE for TKIT-75
        "GTD temperature"
    )
    GTDpressure2 = "GTD pressure - sensor 2"  # change enums to UPPER_CASE for TKIT-75
    GTDtemperature2 = "GTD temperature - sensor 2"  # change enums to UPPER_CASE for TKIT-75
    optodeOxygen = (  # change enums to UPPER_CASE for TKIT-75
        "optode oxygen"
    )
    SBE63phase = "SBE63 oxygen phase"  # change enums to UPPER_CASE for TKIT-75
    SBE63temperature = "SBE63 oxygen temperature"  # change enums to UPPER_CASE for TKIT-75
    dateTime = "date time"  # change enums to UPPER_CASE for TKIT-75
    # NMEA Devices
    nmeaTime = (  # change enums to UPPER_CASE for TKIT-75
        "NMEA Date Time"
    )
    nmeaLatitude = (  # change enums to UPPER_CASE for TKIT-75
        "NMEA Latitude"
    )
    nmeaLongitude = (  # change enums to UPPER_CASE for TKIT-75
        "NMEA Longitude"
    )
    nmeaDepth = "nmea depth"
    statusAndSign = (  # change enums to UPPER_CASE for TKIT-75
        "status and sign"
    )
    vrsInternal = (  # change enums to UPPER_CASE for TKIT-75
        "vrs internal"
    )
    vrsExternal = (  # change enums to UPPER_CASE for TKIT-75
        "vrs external"
    )
    pHtemperature = (  # change enums to UPPER_CASE for TKIT-75
        "ph temperature"
    )
    vk = "vk"  # change enums to UPPER_CASE for TKIT-75
    ib = "ib"  # change enums to UPPER_CASE for TKIT-75
    ik = "ik"  # change enums to UPPER_CASE for TKIT-75
    relativeHumidity = (  # change enums to UPPER_CASE for TKIT-75
        "relative humidity"
    )
    internalTemperature = (  # change enums to UPPER_CASE for TKIT-75
        "internal temperature"
    )
    errorFlag = (  # change enums to UPPER_CASE for TKIT-75
        "error flag"
    )
    surfacePAR = "surface par"
    SBE911PumpStatus = "SBE911 pump status"
    SBE911BottomContactStatus = "SBE911 bottom contact status"
    SBE911ConfirmStatus = "SBE911 confirm status"
    SBE911ModemStatus = "SBE911 modem status"
    dataIntegrity = "data integrity"
    systemTime = "system time"


# export type HexDataTypeStrings = keyof typeof HexDataTypes;

HEX_LENGTH = {
    "temperature": 6,
    "conductivity": 6,
    "pressure": 6,
    "temperatureCompensation": 4,
    "voltage": 4,
    "SBE911Voltage": 3,
    "SBE911SPAR": 3,
    "SBE911TemperatureCompensation": 3,
    "SBE911Status": 1,
    "SBE911DataIntegrity": 2,
    "SBE38temperature": 6,
    "wetlabsSingleSensor": 4,  # There are three of these for each WL Sensor
    "GTDpressure": 8,
    "GTDtemperature": 6,
    "optodeOxygen": 6,
    "SBE63phase": 6,
    "SBE63temperature": 6,
    "SeaFETVint": 6,
    "SeaFETVext": 6,
    "SeaOWLChannel": 4,  # There are three of these for each SeaOWL
    "time": 8,
    "vrsExternal": 6,
    "vrsInternal": 6,
    "pHtemperature": 6,
    "vk": 6,
    "ib": 6,
    "ik": 6,
    "relativeHumidity": 3,
    "internalTemperature": 3,
    "errorFlag": 4,
    # nmea devices
    "nmeaLatitude": 6,
    "nmeaLongitude": 6,
    "nmeaTime": 8,
    "nmeaLocation": 14,
    "statusAndSign": 2,
    "systemTime": 8,
}


class Sensors(Enum):
    """Available sensors to read hex data from"""

    Temperature = (  # change enums to UPPER_CASE for TKIT-75
        "Temperature"
    )
    SecondaryTemperature = "SecondaryTemperature"  # change enums to UPPER_CASE for TKIT-75
    Conductivity = (  # change enums to UPPER_CASE for TKIT-75
        "Conductivity"
    )
    SecondaryConductivity = "SecondaryConductivity"  # change enums to UPPER_CASE for TKIT-75
    Pressure = "Pressure"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt0 = "ExtVolt0"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt1 = "ExtVolt1"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt2 = "ExtVolt2"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt3 = "ExtVolt3"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt4 = "ExtVolt4"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt5 = "ExtVolt5"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt6 = "ExtVolt6"  # change enums to UPPER_CASE for TKIT-75
    ExtVolt7 = "ExtVolt7"  # change enums to UPPER_CASE for TKIT-75
    WETLABS = "WETLABS"  # change enums to UPPER_CASE for TKIT-75
    GTD = "GTD"  # change enums to UPPER_CASE for TKIT-75
    DualGTD = "DualGTD"  # change enums to UPPER_CASE for TKIT-75
    OPTODE = "OPTODE"  # change enums to UPPER_CASE for TKIT-75
    SBE63 = "SBE63"  # change enums to UPPER_CASE for TKIT-75
    SBE38 = "SBE38"  # change enums to UPPER_CASE for TKIT-75
    SeaFET = "SeaFET"  # change enums to UPPER_CASE for TKIT-75
    SPAR = "SPAR"  # change enums to UPPER_CASE for TKIT-75
    # nmea devices
    nmeaLatitude = "nmeaLatitude"  # change enums to UPPER_CASE for TKIT-75
    nmeaLongitude = (  # change enums to UPPER_CASE for TKIT-75
        "nmeaLongitude"
    )
    statusAndSign = (  # change enums to UPPER_CASE for TKIT-75
        "StatusAndSign"
    )
    nmeaTime = "nmeaTime"  # change enums to UPPER_CASE for TKIT-75
    nmeaLocation = "nmeaLocation"  # change enums to UPPER_CASE for TKIT-75
    nmeaDepth = "nmeaDepth"  # change enums to UPPER_CASE for TKIT-75
    SystemTime = "systemTime"  # change enums to UPPER_CASE for TKIT-75


@dataclass
class MeasurementSeries:
    """Container for measurement data."""

    label: str
    description: str
    units: str
    start_time: Optional[date]
    values: np.ndarray


@dataclass
class InstrumentData:
    """Container for instrument data parsed from a CNV file."""

    measurements: Dict[str, MeasurementSeries]
    """
    Dictionary of MeasurementSeries by their label.

    Note: duplicate labels can occur and cnv_to_instrument_data will numerically
    increment the duplicate labels, e.g. the second "depSM" becomes "depSM1".
    """
    interval_s: Optional[float]
    latitude: float
    start_time: Optional[date]
    sample_count: Optional[int]

    def _to_dataframe(self):
        measurements = {k: v.values for k, v in self.measurements.items()}
        return pd.DataFrame(measurements)


def cnv_to_instrument_data(filepath: Union[Path, str]) -> InstrumentData:
    """
    Import the data from a .cnv file and put it into an InstrumentData object.

    Duplicate labels will be incremented for InstrumentData.measurements keys.
    For example, the second "depSM" becomes "depSM1".
    However, the MeasurementSeries.label will be the original label.

    :param filepath: the path to the .cnv file to be imported

    :return: the imported data from the .cnv file

    """

    data = InstrumentData(
        measurements={},
        interval_s=None,
        latitude=0.0,
        start_time=None,
        sample_count=None,
    )

    n = 0

    logger.info("Unpacking instrument data from file: %s", filepath)

    with open(filepath, mode="r") as cnv:
        for line in cnv:
            if line.startswith("*") or line.startswith("#"):
                if line.startswith("# nvalues = "):
                    data.sample_count = int(line[line.find("= ") + 2 : line.find("\n")])
                elif line.startswith("# name "):
                    label = line[line.find("= ") + 2 : line.find(":")]
                    left_bracket = line.find(" [")
                    if left_bracket > 0:
                        description = line[line.find(": ") + 2 : left_bracket]
                        units = line[line.find("[") + 1 : line.find("]")]
                    else:
                        description = line[line.find(": ") + 2 : line.find("\n")].strip()
                        units = ""

                    num_values = data.sample_count or 0  # num_values to 0 if sample_count is None

                    key = label
                    # check for label collision
                    # if collision, increment until find available key
                    i = 1
                    while key in data.measurements:
                        key = f"{label}_{i}"
                        i += 1

                    data.measurements[key] = MeasurementSeries(
                        label=label,
                        description=description,
                        units=units,
                        start_time=None,
                        values=np.zeros(num_values),
                    )

                    if key != label:
                        logger.warning(
                            'duplicate measurement "%s" will use key "%s" in measurements dict',
                            label,
                            key,
                        )

                elif line.startswith("# interval = "):
                    interval = float(
                        line[line.find(": ") + 2 : line.find("\n")]
                    )  # TODO: fix for minutes, hours, etc
                    data.interval_s = interval

                elif line.startswith("# start_time = "):
                    end = min(line.find("\n"), line.find(" ["))
                    date_string = line[line.find("= ") + 2 : end]
                    start_time = datetime.strptime(date_string, "%b %d %Y %H:%M:%S")
                    data.start_time = start_time
                    for measurement in data.measurements.values():
                        measurement.start_time = start_time

                elif line.startswith("** Latitude: "):
                    latitude_parts = line[line.find(": ") + 2 :].split()
                    data.latitude = float(latitude_parts[0]) + float(latitude_parts[1]) / 60.0
                    # TODO: add higher priority latitude to individual measurement series
                    # where necessary

            else:
                values = fix_exponents(" -".join(line.split("-")).split())

                if len(values) > 0:
                    for value, measurement in zip(values, data.measurements.values()):
                        measurement.values[n] = float(value)
                    n += 1
    return data


def fix_exponents(values: List[str]) -> List[str]:
    """Fixes flag values and other numbers with negative exponents.
    This is necessary because sometimes there is only a minus sign
    separating two values in a cnv file. So we split values on the minus
    sign which also splits negative exponents (e.g. 1e-2 becomes 1e, -2).
    This function repairs the exponents by merging numbers that end in
    'e' with the following number in the list (e.g. 1e, -2 becomes 1e-2),
    then removes the extra exponent from the list.

    :param values: List of strings representing numbers

    :return: List of strings where eponents have been fixed
    """

    del_indices = [n + 1 for n, value in enumerate(values) if value.endswith("e")]
    for n in del_indices:
        values[n - 1] = f"{values[n - 1]}{values[n]}"
    new_values = list(np.delete(values, del_indices))
    return new_values


def read_hex_file(
    filepath: Union[Path, str],
    instrument_type: InstrumentType,
    enabled_sensors: List[Sensors] = [],
    moored_mode=False,
    is_shallow=True,
    frequency_channels_suppressed=0,
    voltage_words_suppressed=0,
) -> pd.DataFrame:
    """Reads a .hex file

    :param filepath: path to the .hex file
    :param instrument_type: the instrument that generated the .hex file
    :param enabled_sensors: list of sensors that were enabled on the
        instrument
    :param moored_mode: whether the instrument was in moored or profiling
        mode, defaults to False
    :return: a pandas DataFrame with the hex data
    """
    data_count = 0
    is_data = False

    # iterating over file twice in order to preallocate arrays
    # TODO: Fix this
    file = open(filepath, mode="r")
    for line in file:
        if is_data and not (line == "" or line.startswith("\n") or line.startswith("\r")):
            data_count += 1
        if line.startswith("*END*"):
            is_data = True

    data_length = data_count
    file.seek(0)
    data = pd.DataFrame()
    data_count = 0
    is_data = False

    for line in file:
        if is_data and not (line == "" or line.startswith("\n") or line.startswith("\r")):
            if data_count == 0:
                data = preallocate_dataframe(
                    instrument_type, line, enabled_sensors, moored_mode, data_length
                )
            hex_data = read_hex(
                instrument_type,
                line,
                enabled_sensors,
                moored_mode,
                is_shallow,
                frequency_channels_suppressed,
                voltage_words_suppressed,
            )
            for key, value in hex_data.items():
                data.loc[data_count, key] = value
            data_count += 1
        if line.startswith("*END*"):
            is_data = True

    file.close()

    return data


def preallocate_dataframe(
    instrument_type: InstrumentType,
    line: str,
    enabled_sensors: List[Sensors],
    moored_mode: bool,
    data_length: int,
) -> pd.DataFrame:
    """Prefills a pandas DataFrame with zeros for the instrument data

    :param instrument_type: the instrument type
    :param line: TODO: remove in TKIT-63
    :param enabled_sensors: list of sensors that were enabled on the
        instrument
    :param moored_mode: whether the 19 plus was in moored or profiling
        mode
    :param data_length: the number of rows of data in the hex file

    :return: a dataframe fill of zeros
    """
    sensors = {}
    hex_data = read_hex(instrument_type, line, enabled_sensors, moored_mode)
    # hex_keys = pd.DataFrame(hex_data, index=[0]).columns
    for key, value in hex_data.items():
        if isinstance(value, datetime):
            sensors[key] = pd.date_range(start="2000-01-01", end="2000-01-02", periods=data_length)
        else:
            sensors[key] = np.zeros(data_length)

    return pd.DataFrame(sensors)


def read_hex(
    instrument_type: InstrumentType,
    hex_segment: str = "",
    enabled_sensors: List[Sensors] | None = None,
    moored_mode=False,
    is_shallow=True,
    frequency_channels_suppressed=0,
    voltage_words_suppressed=0,
    hex=hex,
) -> dict:
    """Converts an instrument data hex string into engineering units.

    :param instrument_type: determines how units are converted
    :param hex_segment: one line from a hex data file
    :param enabled_sensors: mooredMode parses time for 19plus in moored
        mode if true
    :param moored_mode: array of Sensors that are enabled. For 37 this
        is always temperature, conductivity, pressure. Defaults to False
    :param hex: Deprecated, use hex_segment

    :return: the sensor values in engineering units that were extracted
        from the input hex string
    """
    if hex is not builtins.hex:
        warnings.warn("hex is deprecated, use hex_segment", DeprecationWarning)

    if instrument_type == InstrumentType.SBE19Plus:
        return read_SBE19plus_format_0(hex_segment, enabled_sensors, moored_mode)

    elif instrument_type == InstrumentType.SBE16Plus:
        return read_SBE19plus_format_0(hex_segment, enabled_sensors, True)

    elif instrument_type in [
        InstrumentType.SBE37IM,
        InstrumentType.SBE37IMP,
        InstrumentType.SBE37IMPODO,
        InstrumentType.SBE37IM,
        InstrumentType.SBE37SI,
        InstrumentType.SBE37SIP,
        InstrumentType.SBE37SM,
        InstrumentType.SBE37SMP,
        InstrumentType.SBE37SMPODO,
        InstrumentType.HydroCAT,
        InstrumentType.HydroCATODO,
    ]:
        return read_SBE37SM_format_0(hex_segment, enabled_sensors)

    elif instrument_type in [InstrumentType.SBE39Plus, InstrumentType.SBE39PlusIM]:
        return read_SBE39plus_format_0(hex_segment, enabled_sensors)

    elif instrument_type in [InstrumentType.SeaFET2, InstrumentType.SeapHox2]:
        return read_seafet_format_0(hex_segment, instrument_type, is_shallow)

    elif instrument_type == InstrumentType.SBE911Plus:
        return read_SBE911plus_format_0(
            hex_segment,
            enabled_sensors,
            frequency_channels_suppressed,
            voltage_words_suppressed,
        )

    else:
        raise ValueError(f"Unsupported instrument type: {instrument_type}")

    return {}


def read_SBE39plus_format_0(
    hex_segment: str = "",
    enabled_sensors: List[Sensors] | None = None,
    hex=builtins.hex,
) -> Dict[str, Union[int, float, datetime]]:
    """Converts a 39plus data hex string into engineering units.

    :param hex_segment: one line from a hex data file
    :param enabled_sensors: array of Sensors that are enabled
    :param hex: Deprecated, use hex_segment

    :return: the 39plus sensor values in engineering units that were
        extracted from the input hex string
    """
    if hex is not builtins.hex:
        warnings.warn("hex is deprecated, use hex_segment", DeprecationWarning)

    results: Dict[str, Union[int, float, datetime]] = {}
    n = 0

    # Datetime (naive, no timezone conversion)
    seconds_since_2000 = int(hex_segment[n : n + HEX_LENGTH["time"]], 16)
    results[HexDataTypes.dateTime.value] = datetime(1970, 1, 1) + timedelta(
        seconds=seconds_since_2000 + SECONDS_BETWEEN_EPOCH_AND_2000
    )
    n += HEX_LENGTH["time"]

    # Temperature
    results[HexDataTypes.temperature.value] = int(
        hex_segment[n : n + HEX_LENGTH["temperature"]], 16
    )
    n += HEX_LENGTH["temperature"]

    # Pressure and temperature compensation
    if enabled_sensors and Sensors.Pressure in enabled_sensors:
        results[HexDataTypes.pressure.value] = int(hex_segment[n : n + HEX_LENGTH["pressure"]], 16)
        n += HEX_LENGTH["pressure"]

        temp_comp = int(hex_segment[n : n + HEX_LENGTH["temperatureCompensation"]], 16)
        results[HexDataTypes.temperatureCompensation.value] = temp_comp
        n += HEX_LENGTH["temperatureCompensation"]

    # Validate hex length
    if n != len(hex_segment.strip()):
        raise ValueError("Hex string length does not match expectation based on enabled sensors")

    return results


def read_seafet_format_0(
    hex_segment: str,
    instrument_type: InstrumentType,
    is_shallow: bool = True,
    hex=builtins.hex,
) -> Dict[str, Union[int, float, datetime]]:
    """Converts a SeaFET2 or SeapHox2 hex string into engineering units.

    :param hex_segment: one line from a hex data file
    :param instrument_type: InstrumentType.SeaFET2 or InstrumentType.SeapHox2
    :param is_shallow: if True, include internal pH and pH reference temperature
    :param hex: Deprecated, use hex_segment

    :return: sensor values in engineering units extracted from the hex string
    """
    if hex is not builtins.hex:
        warnings.warn("hex is deprecated, use hex_segment", DeprecationWarning)

    if instrument_type not in (InstrumentType.SeaFET2, InstrumentType.SeapHox2):
        raise ValueError(
            f"In read_seafet_format_0 {instrument_type} is not recognized as a SeaFET2 or SeapHox2"
        )

    results: Dict[str, Union[int, float, datetime]] = {}
    n = 0

    # SeapHox2 specific values
    if instrument_type == InstrumentType.SeapHox2:
        results[HexDataTypes.temperature.value] = int(
            hex_segment[n : n + HEX_LENGTH["temperature"]], 16
        )
        n += HEX_LENGTH["temperature"]

        results[HexDataTypes.conductivity.value] = (
            int(hex_segment[n : n + HEX_LENGTH["conductivity"]], 16) / 256
        )
        n += HEX_LENGTH["conductivity"]

        results[HexDataTypes.pressure.value] = int(hex_segment[n : n + HEX_LENGTH["pressure"]], 16)
        n += HEX_LENGTH["pressure"]

        results[HexDataTypes.temperatureCompensation.value] = int(
            hex_segment[n : n + HEX_LENGTH["temperatureCompensation"]], 16
        )
        n += HEX_LENGTH["temperatureCompensation"]

        results[HexDataTypes.SBE63phase.value] = (
            int(hex_segment[n : n + HEX_LENGTH["SBE63phase"]], 16) / 100000 - 10
        )
        n += HEX_LENGTH["SBE63phase"]

        results[HexDataTypes.SBE63temperature.value] = (
            int(hex_segment[n : n + HEX_LENGTH["SBE63temperature"]], 16) / 1000000 - 1
        )
        n += HEX_LENGTH["SBE63temperature"]

    # External pH
    results[HexDataTypes.vrsExternal.value] = int(
        hex_segment[n : n + HEX_LENGTH["vrsExternal"]], 16
    )
    n += HEX_LENGTH["vrsExternal"]

    if is_shallow:
        # Internal pH
        results[HexDataTypes.vrsInternal.value] = int(
            hex_segment[n : n + HEX_LENGTH["vrsInternal"]], 16
        )
        n += HEX_LENGTH["vrsInternal"]

        # pH reference temperature
        results[HexDataTypes.pHtemperature.value] = int(
            hex_segment[n : n + HEX_LENGTH["pHtemperature"]], 16
        )
        n += HEX_LENGTH["pHtemperature"]

    # Other sensors
    results[HexDataTypes.vk.value] = int(hex_segment[n : n + HEX_LENGTH["vk"]], 16)
    n += HEX_LENGTH["vk"]

    results[HexDataTypes.ib.value] = int(hex_segment[n : n + HEX_LENGTH["ib"]], 16)
    n += HEX_LENGTH["ib"]

    results[HexDataTypes.ik.value] = int(hex_segment[n : n + HEX_LENGTH["ik"]], 16)
    n += HEX_LENGTH["ik"]

    results[HexDataTypes.relativeHumidity.value] = int(
        hex_segment[n : n + HEX_LENGTH["relativeHumidity"]] + "0", 16
    )
    n += HEX_LENGTH["relativeHumidity"]

    results[HexDataTypes.internalTemperature.value] = int(
        hex_segment[n : n + HEX_LENGTH["internalTemperature"]] + "0", 16
    )
    n += HEX_LENGTH["internalTemperature"]

    # Datetime (naive, no timezone conversion)
    seconds_since_2000 = int(hex_segment[n : n + HEX_LENGTH["time"]], 16)
    results[HexDataTypes.dateTime.value] = datetime(2000, 1, 1) + timedelta(
        seconds=seconds_since_2000
    )
    n += HEX_LENGTH["time"]

    # Error flag
    results[HexDataTypes.errorFlag.value] = int(hex_segment[n : n + HEX_LENGTH["errorFlag"]], 16)
    n += HEX_LENGTH["errorFlag"]

    # Validate hex length
    if n != len(hex_segment.strip()):
        raise ValueError(
            "Hex string length does not match expectation based on instrument_type and is_shallow"
        )

    return results


# TODO: change this to be snake_case for TKIT-75
def read_SBE911plus_format_0(
    hex_segment: str = "",
    enabled_sensors: list["Sensors"] | None = None,
    frequency_channels_suppressed: int = 0,
    voltage_words_suppressed: int = 0,
) -> dict[str, Union[int, float, datetime]]:
    """Converts a 911Plus hex string into engineering units.

    :param hex_segment: one line from a hex data file
    :param enabled_sensors: list of enabled Sensors
    :param frequency_channels_suppressed: number of suppressed frequency channels
    :param voltage_words_suppressed: number of suppressed voltage words

    :return: dictionary of sensor values in engineering units
    """
    if enabled_sensors is None:
        enabled_sensors = []

    results: dict[str, Union[int, float, datetime]] = {}
    n = 0

    # Temperature
    results[HexDataTypes.temperature.value] = frequency_from_3_bytes(
        hex_segment[n : n + HEX_LENGTH["temperature"]]
    )
    n += HEX_LENGTH["temperature"]

    # Conductivity
    results[HexDataTypes.conductivity.value] = frequency_from_3_bytes(
        hex_segment[n : n + HEX_LENGTH["conductivity"]]
    )
    n += HEX_LENGTH["conductivity"]

    # Digiquartz Pressure
    results[HexDataTypes.digiquartzPressure.value] = frequency_from_3_bytes(
        hex_segment[n : n + HEX_LENGTH["pressure"]]
    )
    n += HEX_LENGTH["pressure"]

    # Secondary temperature
    if Sensors.SecondaryTemperature in enabled_sensors:
        results[HexDataTypes.secondaryTemperature.value] = frequency_from_3_bytes(
            hex_segment[n : n + HEX_LENGTH["temperature"]]
        )
        n += HEX_LENGTH["temperature"]
    elif frequency_channels_suppressed <= 1:
        n += HEX_LENGTH["temperature"]

    # Secondary conductivity
    if Sensors.SecondaryConductivity in enabled_sensors:
        results[HexDataTypes.secondaryConductivity.value] = frequency_from_3_bytes(
            hex_segment[n : n + HEX_LENGTH["conductivity"]]
        )
        n += HEX_LENGTH["conductivity"]
    elif frequency_channels_suppressed == 0:
        n += HEX_LENGTH["conductivity"]

    # Voltage channels (suppressed in pairs)
    if Sensors.ExtVolt0 in enabled_sensors or Sensors.ExtVolt1 in enabled_sensors:
        results[HexDataTypes.ExtVolt0.value], results[HexDataTypes.ExtVolt1.value] = (
            voltages_from_3_bytes(hex_segment[n : n + HEX_LENGTH["SBE911Voltage"] * 2])
        )
        n += HEX_LENGTH["SBE911Voltage"] * 2
    elif voltage_words_suppressed <= 3:
        # NotInUse volt channel that was not suppressed
        n += HEX_LENGTH["SBE911Voltage"] * 2

    if Sensors.ExtVolt2 in enabled_sensors or Sensors.ExtVolt3 in enabled_sensors:
        results[HexDataTypes.ExtVolt2.value], results[HexDataTypes.ExtVolt3.value] = (
            voltages_from_3_bytes(hex_segment[n : n + HEX_LENGTH["SBE911Voltage"] * 2])
        )
        n += HEX_LENGTH["SBE911Voltage"] * 2
    elif voltage_words_suppressed <= 2:
        # NotInUse volt channel that was not suppressed
        n += HEX_LENGTH["SBE911Voltage"] * 2

    if Sensors.ExtVolt4 in enabled_sensors or Sensors.ExtVolt5 in enabled_sensors:
        results[HexDataTypes.ExtVolt4.value], results[HexDataTypes.ExtVolt5.value] = (
            voltages_from_3_bytes(hex_segment[n : n + HEX_LENGTH["SBE911Voltage"] * 2])
        )
        n += HEX_LENGTH["SBE911Voltage"] * 2
    elif voltage_words_suppressed <= 1:
        # NotInUse volt channel that was not suppressed
        n += HEX_LENGTH["SBE911Voltage"] * 2

    if Sensors.ExtVolt6 in enabled_sensors or Sensors.ExtVolt7 in enabled_sensors:
        results[HexDataTypes.ExtVolt6.value], results[HexDataTypes.ExtVolt7.value] = (
            voltages_from_3_bytes(hex_segment[n : n + HEX_LENGTH["SBE911Voltage"] * 2])
        )
        n += HEX_LENGTH["SBE911Voltage"] * 2
    elif voltage_words_suppressed == 0:
        # NotInUse volt channel that was not suppressed
        n += HEX_LENGTH["SBE911Voltage"] * 2

    # Surface PAR
    if Sensors.SPAR in enabled_sensors:
        n += 3  # unused bits
        results[HexDataTypes.surfacePAR.value] = (
            int(hex_segment[n : n + HEX_LENGTH["SBE911SPAR"]], 16) / 819
        )
        n += HEX_LENGTH["SBE911SPAR"]

    # NMEA Location
    if Sensors.nmeaLocation in enabled_sensors:
        hex_loc = hex_segment[n : n + HEX_LENGTH["nmeaLocation"]]

        def get_lat_lon(hex_segment: str) -> tuple[float, float]:
            byte0 = int(hex_segment[0:2], 16)
            byte1 = int(hex_segment[2:4], 16)
            byte2 = int(hex_segment[4:6], 16)
            byte3 = int(hex_segment[6:8], 16)
            byte4 = int(hex_segment[8:10], 16)
            byte5 = int(hex_segment[10:12], 16)
            byte6 = format(int(hex_segment[12:14], 16), "08b")
            lat_sign = 1 if byte6[0] == "0" else -1
            lon_sign = 1 if byte6[1] == "0" else -1
            lat = lat_sign * (byte0 * 65536 + byte1 * 256 + byte2) / 50000
            lon = lon_sign * (byte3 * 65536 + byte4 * 256 + byte5) / 50000
            return lat, lon

        (
            results[HexDataTypes.nmeaLatitude.value],
            results[HexDataTypes.nmeaLongitude.value],
        ) = get_lat_lon(hex_loc)
        n += HEX_LENGTH["nmeaLocation"]

    # NMEA Depth
    if Sensors.nmeaDepth in enabled_sensors:
        results[HexDataTypes.nmeaDepth.value] = int(
            hex_segment[n : n + HEX_LENGTH["nmeaDepth"]], 16
        )
        n += HEX_LENGTH["nmeaDepth"]

    # NMEA Time
    if Sensors.nmeaTime in enabled_sensors:
        seconds_since_2000 = int(
            reverse_hex_bytes(hex_segment[n : n + HEX_LENGTH["nmeaTime"]]), 16
        )
        results[HexDataTypes.nmeaTime.value] = datetime(2000, 1, 1) + timedelta(
            seconds=seconds_since_2000
        )
        n += HEX_LENGTH["nmeaTime"]

    # Temperature compensation
    results[HexDataTypes.temperatureCompensation.value] = int(
        hex_segment[n : n + HEX_LENGTH["SBE911TemperatureCompensation"]], 16
    )
    n += HEX_LENGTH["SBE911TemperatureCompensation"]

    # Status bits
    status_bin = format(int(hex_segment[n : n + HEX_LENGTH["SBE911Status"]], 16), "04b")
    results[HexDataTypes.SBE911PumpStatus.value] = int(status_bin[0])
    results[HexDataTypes.SBE911BottomContactStatus.value] = int(status_bin[1])
    results[HexDataTypes.SBE911ConfirmStatus.value] = int(status_bin[2])
    results[HexDataTypes.SBE911ModemStatus.value] = int(status_bin[3])
    n += HEX_LENGTH["SBE911Status"]

    # Data integrity
    results[HexDataTypes.dataIntegrity.value] = int(
        hex_segment[n : n + HEX_LENGTH["SBE911DataIntegrity"]], 16
    )
    n += HEX_LENGTH["SBE911DataIntegrity"]

    # System time
    if Sensors.SystemTime in enabled_sensors:
        seconds_since_1970 = int(
            reverse_hex_bytes(hex_segment[n : n + HEX_LENGTH["systemTime"]]), 16
        )
        results[HexDataTypes.systemTime.value] = datetime(1970, 1, 1) + timedelta(
            seconds=seconds_since_1970
        )
        n += HEX_LENGTH["systemTime"]

    # Validate hex length
    if n != len(hex_segment.strip()):
        raise ValueError("Hex string length does not match expectation based on enabled sensors")

    return results


# TODO: change the following fn name to be snake_case for TKIT-75
# TODO: Fix these
def read_SBE19plus_format_0(
    hex_segment: str = "",
    enabled_sensors: List[Sensors] | None = None,
    moored_mode=False,
    hex=hex,
) -> Dict[str, Union[float, datetime]]:
    """Converts a 19plus V2 data hex string into engineering units.

    :param hex_segment: one line from a hex data file
    :param enabled_sensors: array of Sensors that are enabled. For 37
        this is always temperature, conductivity, pressure. Defaults to
        False
    :param moored_mode: parses time for 19plus in moored mode if true
    :param hex: Deprecated, use hex_segment

    :return: the 19plus V2 sensor values in engineering units that were
            extracted from the input hex string

    :raises RuntimeWarning: if the hex string length does not match the
        expected length
    """
    if hex is not builtins.hex:
        warnings.warn("hex is deprecated, use hex_segment", DeprecationWarning)

    results: Dict[str, Union[int, float, datetime]] = {}
    n = 0
    for sensor in Sensors:
        if enabled_sensors and sensor in enabled_sensors:
            if sensor == Sensors.Temperature:
                results[HexDataTypes.temperature.value] = int(
                    hex_segment[n : HEX_LENGTH["temperature"]], 16
                )
                n += HEX_LENGTH["temperature"]

            if sensor == Sensors.Conductivity:
                results[HexDataTypes.conductivity.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["conductivity"]], 16) / 256
                )
                n += HEX_LENGTH["conductivity"]

            if sensor == Sensors.Pressure:  # TODO: add conversion for quartz pressure sensors
                results[HexDataTypes.pressure.value] = int(
                    hex_segment[n : n + HEX_LENGTH["pressure"]], 16
                )
                n += HEX_LENGTH["pressure"]
                result = (
                    int(hex_segment[n : n + HEX_LENGTH["temperatureCompensation"]], 16)
                    / COUNTS_TO_VOLTS
                )
                results[HexDataTypes.temperatureCompensation.value] = result
                n += HEX_LENGTH["temperatureCompensation"]

            if sensor in [
                Sensors.ExtVolt0,
                Sensors.ExtVolt1,
                Sensors.ExtVolt2,
                Sensors.ExtVolt3,
                Sensors.ExtVolt4,
                Sensors.ExtVolt5,
            ]:
                result = int(hex_segment[n : n + HEX_LENGTH["voltage"]], 16) / COUNTS_TO_VOLTS
                results[HexDataTypes[Sensors[sensor.value].value].value] = result
                n += HEX_LENGTH["voltage"]

            if sensor == Sensors.SBE38:
                results[HexDataTypes.SBE38temperature.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["SBE38temperature"]], 16) / 100000 - 10
                )
                n += HEX_LENGTH["SBE38temperature"]

            if sensor == Sensors.WETLABS:
                results[HexDataTypes.wetlabs0.value] = int(
                    hex_segment[n : n + HEX_LENGTH["wetlabsSingleSensor"]], 16
                )
                n += HEX_LENGTH["wetlabsSingleSensor"]

                results[HexDataTypes.wetlabs1.value] = int(
                    hex_segment[n : n + HEX_LENGTH["wetlabsSingleSensor"]], 16
                )
                n += HEX_LENGTH["wetlabsSingleSensor"]

                results[HexDataTypes.wetlabs2.value] = int(
                    hex_segment[n : n + HEX_LENGTH["wetlabsSingleSensor"]], 16
                )
                n += HEX_LENGTH["wetlabsSingleSensor"]

            if sensor == Sensors.GTD:
                results[HexDataTypes.GTDpressure.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["GTDpressure"]], 16) / 10000
                )
                n += HEX_LENGTH["GTDpressure"]
                results[HexDataTypes.GTDtemperature.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["GTDtemperature"]], 16) / 10000 - 10
                )
                n += HEX_LENGTH["GTDtemperature"]

            if sensor == Sensors.DualGTD:
                results[HexDataTypes.GTDpressure.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["GTDpressure"]], 16) / 10000
                )
                n += HEX_LENGTH["GTDpressure"]
                results[HexDataTypes.GTDtemperature.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["GTDtemperature"]], 16) / 10000 - 10
                )
                n += HEX_LENGTH["GTDtemperature"]
                results[HexDataTypes.GTDpressure2.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["GTDpressure"]], 16) / 10000
                )
                n += HEX_LENGTH["GTDpressure"]
                results[HexDataTypes.GTDtemperature2.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["GTDtemperature"]], 16) / 10000 - 10
                )
                n += HEX_LENGTH["GTDtemperature"]

            if sensor == Sensors.OPTODE:
                results[HexDataTypes.optodeOxygen.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["optodeOxygen"]], 16) / 10000 - 10
                )
                n += HEX_LENGTH["optodeOxygen"]

            if sensor == Sensors.SBE63:
                results[HexDataTypes.SBE63phase.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["SBE63phase"]], 16) / 100000 - 10
                )
                n += HEX_LENGTH["SBE63phase"]
                results[HexDataTypes.SBE63temperature.value] = (
                    int(hex_segment[n : n + HEX_LENGTH["SBE63temperature"]], 16) / 1000000 - 1
                )
                n += HEX_LENGTH["SBE63temperature"]

            # Extract NMEA Sensors
            if sensor == Sensors.nmeaLatitude:
                lat = read_nmea_coordinates(hex_segment[n : n + HEX_LENGTH["nmeaLatitude"]])
                results[HexDataTypes.nmeaLatitude.value] = lat
                n += HEX_LENGTH["nmeaLatitude"]
            if sensor == Sensors.nmeaLongitude:
                lon = read_nmea_coordinates(hex_segment[n : n + HEX_LENGTH["nmeaLongitude"]])
                results[HexDataTypes.nmeaLongitude.value] = lon
                n += HEX_LENGTH["nmeaLongitude"]
            if sensor == Sensors.statusAndSign:
                signs = read_status_sign(hex_segment[n : n + HEX_LENGTH["statusAndSign"]])
                results[HexDataTypes.nmeaLatitude.value] *= signs[0]
                results[HexDataTypes.nmeaLongitude.value] *= signs[1]
                n += HEX_LENGTH["statusAndSign"]
            if sensor == Sensors.nmeaTime:
                seconds_since_2000 = read_nmea_time(hex_segment[n : n + HEX_LENGTH["nmeaTime"]])
                # naive, no timezone conversion
                results[HexDataTypes.nmeaTime.value] = datetime(1970, 1, 1) + timedelta(
                    seconds=seconds_since_2000 + SECONDS_BETWEEN_EPOCH_AND_2000
                )
                n += HEX_LENGTH["nmeaTime"]

    if moored_mode:
        seconds_since_2000 = int(hex_segment[n : n + HEX_LENGTH["time"]], 16)
        results[HexDataTypes.dateTime.value] = datetime.fromtimestamp(
            seconds_since_2000 + SECONDS_BETWEEN_EPOCH_AND_2000
        )
        n += HEX_LENGTH["time"]

    # Validate hex length. Ensure length matches what is expected based
    # on enabled sensors and moored mode.
    if n != len(hex_segment.strip()):
        raise RuntimeWarning(
            "Hex string length does not match expectation based on enabled sensors and moored mode"
        )

    return results


# TODO: change this to be snake_case for TKIT-75
def read_SBE37SM_format_0(
    hex_segment: str = "",
    enabled_sensors: List[Sensors] | None = None,
    hex=hex,
) -> Dict[str, Union[int, float, datetime]]:
    """Converts a 37 family data hex string into engineering units.

    :param hex_segment: one line from a hex data file
    :param enabled_sensors: array of Sensors that are enabled. For 37
        this is always temperature, conductivity, pressure. Defaults to
        False
    :param hex: Deprecated, use hex_segment

    :return: the 37 family sensor values in engineering units that were
        extracted from the input hex string
    """
    if hex is not builtins.hex:
        warnings.warn("hex is deprecated, use hex_segment", DeprecationWarning)

    results: Dict[str, Union[int, float, datetime]] = {}
    n = 0
    results[HexDataTypes.temperature.value] = int(hex_segment[n : HEX_LENGTH["temperature"]], 16)
    n += HEX_LENGTH["temperature"]

    results[HexDataTypes.conductivity.value] = (
        int(hex_segment[n : n + HEX_LENGTH["conductivity"]], 16) / 256
    )
    n += HEX_LENGTH["conductivity"]

    if enabled_sensors and Sensors.SBE63 in enabled_sensors:
        results[HexDataTypes.SBE63phase.value] = (
            int(hex_segment[n : n + HEX_LENGTH["SBE63phase"]], 16) / 100000 - 10
        )
        n += HEX_LENGTH["SBE63phase"]
        results[HexDataTypes.SBE63temperature.value] = (
            int(hex_segment[n : n + HEX_LENGTH["SBE63temperature"]], 16) / 1000000 - 1
        )
        n += HEX_LENGTH["SBE63temperature"]

    if enabled_sensors and Sensors.Pressure in enabled_sensors:
        results[HexDataTypes.pressure.value] = int(hex_segment[n : n + HEX_LENGTH["pressure"]], 16)
        n += HEX_LENGTH["pressure"]
        result = int(hex_segment[n : n + HEX_LENGTH["temperatureCompensation"]], 16)
        results[HexDataTypes.temperatureCompensation.value] = result
        n += HEX_LENGTH["temperatureCompensation"]

    seconds_since_2000 = int(hex_segment[n : n + HEX_LENGTH["time"]], 16)
    results[HexDataTypes.dateTime.value] = datetime(1970, 1, 1) + timedelta(
        seconds=seconds_since_2000 + SECONDS_BETWEEN_EPOCH_AND_2000
    )
    n += HEX_LENGTH["time"]

    # Validate hex length
    if n != len(hex_segment.strip()):
        raise ValueError("Hex string length does not match expectation based on enabled sensors")

    return results


def read_nmea_coordinates(hex_segment: str):
    """Converts a 3 byte NMEA hex string to latitude or longitude

    :param hex_segment: 3 byte hex string
    :raises RuntimeWarning: raised if the hex string is the wrong length
    :return: latitude or longitide coordinate
    """
    if len(hex_segment) != 6:
        raise RuntimeWarning(
            f"Unknown Coordinate Format. Received Hex of length {len(hex_segment)}. "
            f"Should have received Hex of length {HEX_LENGTH['nmeaLongitude']}"
        )
    byte0 = int(hex_segment[0:2], 16)
    byte1 = int(hex_segment[2:4], 16)
    byte2 = int(hex_segment[4:6], 16)
    coordinate = (byte0 * 65536 + byte1 * 256 + byte2) / 50000
    return coordinate


def read_status_sign(hex_segment: str):
    """Converts a hex byte to the signs for NMEA latitude and longitude

    :param hex_segment: 1 byte hex string
    :raises RuntimeWarning: raised if the hex string is the wrong length
    :raises RuntimeWarning: raised when the signs are converted
        incorrectly
    :return: a list of two integers (1 or -1)
    """
    if len(hex_segment) != 2:
        raise RuntimeWarning("Unknown Status Format")
    integer = int(hex_segment, 16)
    binary = format(integer, "0>8b")
    signs = []
    if binary[0] == "0":
        signs.append(1)
    elif binary[0] == "1":
        signs.append(-1)

    if binary[1] == "0":
        signs.append(1)
    elif binary[1] == "1":
        signs.append(-1)
    if len(signs) != 2:
        raise RuntimeWarning("An error occured while processing Coordinate Signs")
    return signs


def read_nmea_time(hex_segment: str):
    """Convert an 8 byte hex string to the number of seconds since 2000

    :param hex_segment: an 8 byte hex string
    :raises RuntimeWarning: raised if the hex string is the wrong length
    :return: _description_
    """
    if len(hex_segment) != 8:
        raise RuntimeWarning("Unknown Time Format")
    byte0 = hex_segment[0:2]
    byte1 = hex_segment[2:4]
    byte2 = hex_segment[4:6]
    byte3 = hex_segment[6:8]
    reformatted = int(byte3 + byte2 + byte1 + byte0, 16)
    return reformatted


def frequency_from_3_bytes(hex_segment: str) -> float:
    """Convert 3 bytes to a frequency value in Hz.

    :param hex_segment: 3-byte (6-character) hex string
    :return: frequency in Hz
    """
    byte0 = int(hex_segment[0:2], 16)
    byte1 = int(hex_segment[2:4], 16)
    byte2 = int(hex_segment[4:6], 16)
    # SBE11plus V2 user manual version A, page 39
    frequency = byte0 * 256 + byte1 + byte2 / 256
    return frequency


def voltages_from_3_bytes(hex_segment: str) -> tuple[float, float]:
    """Convert 3 bytes to two voltage channels.

    Each voltage channel is 12 bits; adjacent channels share a byte.
    Channels are suppressed in pairs, 3 bytes at a time.

    :param hex_segment: 3-byte (6-character) hex string
    :return: tuple of two voltages (voltageA, voltageB)
    """
    decimal_a = int(hex_segment[0 : HEX_LENGTH["SBE911Voltage"]], 16)
    voltage_a = 5 * (1 - decimal_a / 4095)

    decimal_b = int(hex_segment[HEX_LENGTH["SBE911Voltage"] : HEX_LENGTH["SBE911Voltage"] * 2], 16)
    voltage_b = 5 * (1 - decimal_b / 4095)

    return voltage_a, voltage_b


def reverse_hex_bytes(bytes_str: str) -> str:
    """
    Reverse the ASCII hex byte ordering.

    :param bytes_str: The ASCII hex bytes that need reordering.

    :return: The reordered ASCII hex bytes.
    """
    if not bytes_str:
        raise ValueError("Nothing to reverse")

    if len(bytes_str) % 2 != 0:
        raise ValueError("Hex string length is not even")

    if len(bytes_str) == 2:
        return bytes_str

    swapped = ""
    num_bytes = len(bytes_str) // 2

    for idx in range(num_bytes - 1, -1, -1):
        swapped += bytes_str[idx * 2 : idx * 2 + 2]

    return swapped



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""A collection of utility functions related to the processing of SBS
instrument data.
"""
# Functions:
#   close_enough (np.ndarray, np.ndarray, int, float) -> bool
#   plot (np.ndarray)
#   percent_match (np.ndarray, np.ndarray) -> str

# Native imports

# Third-party imports
import matplotlib.pyplot as plt
import numpy as np

# Sea-Bird imports

# Internal imports
from seabirdscientific.processing import FLAG_VALUE


def close_enough(
    test_values: np.ndarray,
    expected_values: np.ndarray,
    rounding_order: int,
    absolute_tolerance: float,
) -> bool:
    """Compares ndarrays, ignoring differences due to rounding or
    truncating the least significant digit. This is only for comparing
    data to legacy software

    Occasionally a float will be accurate to some decimal, but the next
    lower decimal may toggle it over or under 5, causing it to round
    differently if it's rounded or truncated. This function adds and
    subtracts half of the least significant digit, then compares values
    of a given tolerance to either result

    :param test_values: values derived by this library
    :param expected_values: values derived by SeaSoft
    :param rounding_order: The order that CNV values were rounded to
    :param absolute_tolerance: values must be at least this close after
        adding or subtracting the rounding error

    :return: pass or fail aggregate
    """

    results = np.full(len(test_values), False)
    for n, v in enumerate(test_values):
        results[n] = (
            np.round(v, rounding_order) == expected_values[n]
            or np.isclose(
                v,
                expected_values[n] - 0.5 * 10**-rounding_order,
                rtol=0,
                atol=absolute_tolerance,
            )
            or np.isclose(
                v,
                expected_values[n] + 0.5 * 10**-rounding_order,
                rtol=0,
                atol=absolute_tolerance,
            )
        )
    return bool(np.all(results))


def plot(**kwargs: np.ndarray):
    """Plots a dictionary of ndarrays

    :param kwargs: the dictionary to plot
    """

    _, ax = plt.subplots(figsize=(20, 10))
    for key, value in kwargs.items():
        x = range(len(value))
        ax.plot(x, value, label=key)
        ax.legend()
    plt.show()


def percent_match(x1: np.ndarray, x2: np.ndarray) -> str:
    """Calculates the extent that two arrays match.

    :param x1: first array to be compared
    :param x2: second array to be compared

    :return: a message with the percentage of matching elements
    """

    return f"{100 - (x1 != x2).sum() * 100 / len(x1):0.2f}% match"


def get_tolerance(data: np.ndarray):
    """Checks the first 10 values of an array, gets the longest decimal
    length to the right of the decimal, and returns 1/10^length. Used
    for unit tests so results can be compared with a variable tolerance

    :param data: a numpy array of numbers
    :return: the number of significant digits to the right of the decimal
    """
    decimal_lengths = [0]
    for n in range(len(data)):
        if not np.isnan(data[n]) and data[n] != FLAG_VALUE:
            decimal_lengths.append(len(f"{data[n]}".split(".")[1]))
            if len(decimal_lengths) >= 10:
                break
    return 1 / 10 ** max(decimal_lengths)
